# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc
import warnings

import qbitcoin.generated.qbitwallet_pb2 as qbitwallet__pb2

GRPC_GENERATED_VERSION = '1.72.1'
GRPC_VERSION = grpc.__version__
_version_not_supported = False

try:
    from grpc._utilities import first_version_is_lower
    _version_not_supported = first_version_is_lower(GRPC_VERSION, GRPC_GENERATED_VERSION)
except ImportError:
    _version_not_supported = True

if _version_not_supported:
    raise RuntimeError(
        f'The grpc package installed is at version {GRPC_VERSION},'
        + f' but the generated code in qbitwallet_pb2_grpc.py depends on'
        + f' grpcio>={GRPC_GENERATED_VERSION}.'
        + f' Please upgrade your grpc module to grpcio>={GRPC_GENERATED_VERSION}'
        + f' or downgrade your generated code using grpcio-tools<={GRPC_VERSION}.'
    )


class WalletAPIStub(object):
    """//////////////////////////
    //////////////////////////
    //////////////////////////
    ////     API       ///////
    //////////////////////////
    //////////////////////////
    //////////////////////////

    This service describes the Wallet API
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.AddNewAddress = channel.unary_unary(
                '/qrl.WalletAPI/AddNewAddress',
                request_serializer=qbitwallet__pb2.AddNewAddressReq.SerializeToString,
                response_deserializer=qbitwallet__pb2.AddNewAddressResp.FromString,
                _registered_method=True)
        self.AddNewAddressWithSlaves = channel.unary_unary(
                '/qrl.WalletAPI/AddNewAddressWithSlaves',
                request_serializer=qbitwallet__pb2.AddNewAddressWithSlavesReq.SerializeToString,
                response_deserializer=qbitwallet__pb2.AddNewAddressResp.FromString,
                _registered_method=True)
        self.ListAddresses = channel.unary_unary(
                '/qrl.WalletAPI/ListAddresses',
                request_serializer=qbitwallet__pb2.ListAddressesReq.SerializeToString,
                response_deserializer=qbitwallet__pb2.ListAddressesResp.FromString,
                _registered_method=True)
        self.RemoveAddress = channel.unary_unary(
                '/qrl.WalletAPI/RemoveAddress',
                request_serializer=qbitwallet__pb2.RemoveAddressReq.SerializeToString,
                response_deserializer=qbitwallet__pb2.RemoveAddressResp.FromString,
                _registered_method=True)
        self.IsValidAddress = channel.unary_unary(
                '/qrl.WalletAPI/IsValidAddress',
                request_serializer=qbitwallet__pb2.ValidAddressReq.SerializeToString,
                response_deserializer=qbitwallet__pb2.ValidAddressResp.FromString,
                _registered_method=True)
        self.EncryptWallet = channel.unary_unary(
                '/qrl.WalletAPI/EncryptWallet',
                request_serializer=qbitwallet__pb2.EncryptWalletReq.SerializeToString,
                response_deserializer=qbitwallet__pb2.EncryptWalletResp.FromString,
                _registered_method=True)
        self.LockWallet = channel.unary_unary(
                '/qrl.WalletAPI/LockWallet',
                request_serializer=qbitwallet__pb2.LockWalletReq.SerializeToString,
                response_deserializer=qbitwallet__pb2.LockWalletResp.FromString,
                _registered_method=True)
        self.UnlockWallet = channel.unary_unary(
                '/qrl.WalletAPI/UnlockWallet',
                request_serializer=qbitwallet__pb2.UnlockWalletReq.SerializeToString,
                response_deserializer=qbitwallet__pb2.UnlockWalletResp.FromString,
                _registered_method=True)
        self.GetRecoverySeeds = channel.unary_unary(
                '/qrl.WalletAPI/GetRecoverySeeds',
                request_serializer=qbitwallet__pb2.GetRecoverySeedsReq.SerializeToString,
                response_deserializer=qbitwallet__pb2.GetRecoverySeedsResp.FromString,
                _registered_method=True)
        self.GetWalletInfo = channel.unary_unary(
                '/qrl.WalletAPI/GetWalletInfo',
                request_serializer=qbitwallet__pb2.GetWalletInfoReq.SerializeToString,
                response_deserializer=qbitwallet__pb2.GetWalletInfoResp.FromString,
                _registered_method=True)
        self.RelayTransferTxn = channel.unary_unary(
                '/qrl.WalletAPI/RelayTransferTxn',
                request_serializer=qbitwallet__pb2.RelayTransferTxnReq.SerializeToString,
                response_deserializer=qbitwallet__pb2.RelayTxnResp.FromString,
                _registered_method=True)
        self.RelayTransferTxnBySlave = channel.unary_unary(
                '/qrl.WalletAPI/RelayTransferTxnBySlave',
                request_serializer=qbitwallet__pb2.RelayTransferTxnBySlaveReq.SerializeToString,
                response_deserializer=qbitwallet__pb2.RelayTxnResp.FromString,
                _registered_method=True)
        self.RelayMessageTxn = channel.unary_unary(
                '/qrl.WalletAPI/RelayMessageTxn',
                request_serializer=qbitwallet__pb2.RelayMessageTxnReq.SerializeToString,
                response_deserializer=qbitwallet__pb2.RelayTxnResp.FromString,
                _registered_method=True)
        self.RelayMessageTxnBySlave = channel.unary_unary(
                '/qrl.WalletAPI/RelayMessageTxnBySlave',
                request_serializer=qbitwallet__pb2.RelayMessageTxnBySlaveReq.SerializeToString,
                response_deserializer=qbitwallet__pb2.RelayTxnResp.FromString,
                _registered_method=True)
        self.RelayTokenTxn = channel.unary_unary(
                '/qrl.WalletAPI/RelayTokenTxn',
                request_serializer=qbitwallet__pb2.RelayTokenTxnReq.SerializeToString,
                response_deserializer=qbitwallet__pb2.RelayTxnResp.FromString,
                _registered_method=True)
        self.RelayTokenTxnBySlave = channel.unary_unary(
                '/qrl.WalletAPI/RelayTokenTxnBySlave',
                request_serializer=qbitwallet__pb2.RelayTokenTxnBySlaveReq.SerializeToString,
                response_deserializer=qbitwallet__pb2.RelayTxnResp.FromString,
                _registered_method=True)
        self.RelayTransferTokenTxn = channel.unary_unary(
                '/qrl.WalletAPI/RelayTransferTokenTxn',
                request_serializer=qbitwallet__pb2.RelayTransferTokenTxnReq.SerializeToString,
                response_deserializer=qbitwallet__pb2.RelayTxnResp.FromString,
                _registered_method=True)
        self.RelayTransferTokenTxnBySlave = channel.unary_unary(
                '/qrl.WalletAPI/RelayTransferTokenTxnBySlave',
                request_serializer=qbitwallet__pb2.RelayTransferTokenTxnBySlaveReq.SerializeToString,
                response_deserializer=qbitwallet__pb2.RelayTxnResp.FromString,
                _registered_method=True)
        self.RelaySlaveTxn = channel.unary_unary(
                '/qrl.WalletAPI/RelaySlaveTxn',
                request_serializer=qbitwallet__pb2.RelaySlaveTxnReq.SerializeToString,
                response_deserializer=qbitwallet__pb2.RelayTxnResp.FromString,
                _registered_method=True)
        self.RelaySlaveTxnBySlave = channel.unary_unary(
                '/qrl.WalletAPI/RelaySlaveTxnBySlave',
                request_serializer=qbitwallet__pb2.RelaySlaveTxnBySlaveReq.SerializeToString,
                response_deserializer=qbitwallet__pb2.RelayTxnResp.FromString,
                _registered_method=True)
        self.ChangePassphrase = channel.unary_unary(
                '/qrl.WalletAPI/ChangePassphrase',
                request_serializer=qbitwallet__pb2.ChangePassphraseReq.SerializeToString,
                response_deserializer=qbitwallet__pb2.ChangePassphraseResp.FromString,
                _registered_method=True)
        self.GetTransactionsByAddress = channel.unary_unary(
                '/qrl.WalletAPI/GetTransactionsByAddress',
                request_serializer=qbitwallet__pb2.TransactionsByAddressReq.SerializeToString,
                response_deserializer=qbitwallet__pb2.TransactionsByAddressResp.FromString,
                _registered_method=True)
        self.GetPaginatedTransactionsByAddress = channel.unary_unary(
                '/qrl.WalletAPI/GetPaginatedTransactionsByAddress',
                request_serializer=qbitwallet__pb2.PaginatedTransactionsByAddressReq.SerializeToString,
                response_deserializer=qbitwallet__pb2.PaginatedTransactionsByAddressResp.FromString,
                _registered_method=True)
        self.GetTransaction = channel.unary_unary(
                '/qrl.WalletAPI/GetTransaction',
                request_serializer=qbitwallet__pb2.TransactionReq.SerializeToString,
                response_deserializer=qbitwallet__pb2.TransactionResp.FromString,
                _registered_method=True)
        self.GetBalance = channel.unary_unary(
                '/qrl.WalletAPI/GetBalance',
                request_serializer=qbitwallet__pb2.BalanceReq.SerializeToString,
                response_deserializer=qbitwallet__pb2.BalanceResp.FromString,
                _registered_method=True)
        self.GetTotalBalance = channel.unary_unary(
                '/qrl.WalletAPI/GetTotalBalance',
                request_serializer=qbitwallet__pb2.TotalBalanceReq.SerializeToString,
                response_deserializer=qbitwallet__pb2.TotalBalanceResp.FromString,
                _registered_method=True)
        self.GetOTS = channel.unary_unary(
                '/qrl.WalletAPI/GetOTS',
                request_serializer=qbitwallet__pb2.OTSReq.SerializeToString,
                response_deserializer=qbitwallet__pb2.OTSResp.FromString,
                _registered_method=True)
        self.GetHeight = channel.unary_unary(
                '/qrl.WalletAPI/GetHeight',
                request_serializer=qbitwallet__pb2.HeightReq.SerializeToString,
                response_deserializer=qbitwallet__pb2.HeightResp.FromString,
                _registered_method=True)
        self.GetBlock = channel.unary_unary(
                '/qrl.WalletAPI/GetBlock',
                request_serializer=qbitwallet__pb2.BlockReq.SerializeToString,
                response_deserializer=qbitwallet__pb2.BlockResp.FromString,
                _registered_method=True)
        self.GetBlockByNumber = channel.unary_unary(
                '/qrl.WalletAPI/GetBlockByNumber',
                request_serializer=qbitwallet__pb2.BlockByNumberReq.SerializeToString,
                response_deserializer=qbitwallet__pb2.BlockResp.FromString,
                _registered_method=True)
        self.GetAddressFromPK = channel.unary_unary(
                '/qrl.WalletAPI/GetAddressFromPK',
                request_serializer=qbitwallet__pb2.AddressFromPKReq.SerializeToString,
                response_deserializer=qbitwallet__pb2.AddressFromPKResp.FromString,
                _registered_method=True)
        self.GetNodeInfo = channel.unary_unary(
                '/qrl.WalletAPI/GetNodeInfo',
                request_serializer=qbitwallet__pb2.NodeInfoReq.SerializeToString,
                response_deserializer=qbitwallet__pb2.NodeInfoResp.FromString,
                _registered_method=True)


class WalletAPIServicer(object):
    """//////////////////////////
    //////////////////////////
    //////////////////////////
    ////     API       ///////
    //////////////////////////
    //////////////////////////
    //////////////////////////

    This service describes the Wallet API
    """

    def AddNewAddress(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def AddNewAddressWithSlaves(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ListAddresses(self, request, context):
        """rpc AddAddressFromSeed(AddAddressFromSeedReq) returns (AddAddressFromSeedResp);

        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def RemoveAddress(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def IsValidAddress(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def EncryptWallet(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def LockWallet(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def UnlockWallet(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetRecoverySeeds(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetWalletInfo(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def RelayTransferTxn(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def RelayTransferTxnBySlave(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def RelayMessageTxn(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def RelayMessageTxnBySlave(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def RelayTokenTxn(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def RelayTokenTxnBySlave(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def RelayTransferTokenTxn(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def RelayTransferTokenTxnBySlave(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def RelaySlaveTxn(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def RelaySlaveTxnBySlave(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ChangePassphrase(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetTransactionsByAddress(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetPaginatedTransactionsByAddress(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetTransaction(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetBalance(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetTotalBalance(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetOTS(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetHeight(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetBlock(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetBlockByNumber(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetAddressFromPK(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetNodeInfo(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_WalletAPIServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'AddNewAddress': grpc.unary_unary_rpc_method_handler(
                    servicer.AddNewAddress,
                    request_deserializer=qbitwallet__pb2.AddNewAddressReq.FromString,
                    response_serializer=qbitwallet__pb2.AddNewAddressResp.SerializeToString,
            ),
            'AddNewAddressWithSlaves': grpc.unary_unary_rpc_method_handler(
                    servicer.AddNewAddressWithSlaves,
                    request_deserializer=qbitwallet__pb2.AddNewAddressWithSlavesReq.FromString,
                    response_serializer=qbitwallet__pb2.AddNewAddressResp.SerializeToString,
            ),
            'ListAddresses': grpc.unary_unary_rpc_method_handler(
                    servicer.ListAddresses,
                    request_deserializer=qbitwallet__pb2.ListAddressesReq.FromString,
                    response_serializer=qbitwallet__pb2.ListAddressesResp.SerializeToString,
            ),
            'RemoveAddress': grpc.unary_unary_rpc_method_handler(
                    servicer.RemoveAddress,
                    request_deserializer=qbitwallet__pb2.RemoveAddressReq.FromString,
                    response_serializer=qbitwallet__pb2.RemoveAddressResp.SerializeToString,
            ),
            'IsValidAddress': grpc.unary_unary_rpc_method_handler(
                    servicer.IsValidAddress,
                    request_deserializer=qbitwallet__pb2.ValidAddressReq.FromString,
                    response_serializer=qbitwallet__pb2.ValidAddressResp.SerializeToString,
            ),
            'EncryptWallet': grpc.unary_unary_rpc_method_handler(
                    servicer.EncryptWallet,
                    request_deserializer=qbitwallet__pb2.EncryptWalletReq.FromString,
                    response_serializer=qbitwallet__pb2.EncryptWalletResp.SerializeToString,
            ),
            'LockWallet': grpc.unary_unary_rpc_method_handler(
                    servicer.LockWallet,
                    request_deserializer=qbitwallet__pb2.LockWalletReq.FromString,
                    response_serializer=qbitwallet__pb2.LockWalletResp.SerializeToString,
            ),
            'UnlockWallet': grpc.unary_unary_rpc_method_handler(
                    servicer.UnlockWallet,
                    request_deserializer=qbitwallet__pb2.UnlockWalletReq.FromString,
                    response_serializer=qbitwallet__pb2.UnlockWalletResp.SerializeToString,
            ),
            'GetRecoverySeeds': grpc.unary_unary_rpc_method_handler(
                    servicer.GetRecoverySeeds,
                    request_deserializer=qbitwallet__pb2.GetRecoverySeedsReq.FromString,
                    response_serializer=qbitwallet__pb2.GetRecoverySeedsResp.SerializeToString,
            ),
            'GetWalletInfo': grpc.unary_unary_rpc_method_handler(
                    servicer.GetWalletInfo,
                    request_deserializer=qbitwallet__pb2.GetWalletInfoReq.FromString,
                    response_serializer=qbitwallet__pb2.GetWalletInfoResp.SerializeToString,
            ),
            'RelayTransferTxn': grpc.unary_unary_rpc_method_handler(
                    servicer.RelayTransferTxn,
                    request_deserializer=qbitwallet__pb2.RelayTransferTxnReq.FromString,
                    response_serializer=qbitwallet__pb2.RelayTxnResp.SerializeToString,
            ),
            'RelayTransferTxnBySlave': grpc.unary_unary_rpc_method_handler(
                    servicer.RelayTransferTxnBySlave,
                    request_deserializer=qbitwallet__pb2.RelayTransferTxnBySlaveReq.FromString,
                    response_serializer=qbitwallet__pb2.RelayTxnResp.SerializeToString,
            ),
            'RelayMessageTxn': grpc.unary_unary_rpc_method_handler(
                    servicer.RelayMessageTxn,
                    request_deserializer=qbitwallet__pb2.RelayMessageTxnReq.FromString,
                    response_serializer=qbitwallet__pb2.RelayTxnResp.SerializeToString,
            ),
            'RelayMessageTxnBySlave': grpc.unary_unary_rpc_method_handler(
                    servicer.RelayMessageTxnBySlave,
                    request_deserializer=qbitwallet__pb2.RelayMessageTxnBySlaveReq.FromString,
                    response_serializer=qbitwallet__pb2.RelayTxnResp.SerializeToString,
            ),
            'RelayTokenTxn': grpc.unary_unary_rpc_method_handler(
                    servicer.RelayTokenTxn,
                    request_deserializer=qbitwallet__pb2.RelayTokenTxnReq.FromString,
                    response_serializer=qbitwallet__pb2.RelayTxnResp.SerializeToString,
            ),
            'RelayTokenTxnBySlave': grpc.unary_unary_rpc_method_handler(
                    servicer.RelayTokenTxnBySlave,
                    request_deserializer=qbitwallet__pb2.RelayTokenTxnBySlaveReq.FromString,
                    response_serializer=qbitwallet__pb2.RelayTxnResp.SerializeToString,
            ),
            'RelayTransferTokenTxn': grpc.unary_unary_rpc_method_handler(
                    servicer.RelayTransferTokenTxn,
                    request_deserializer=qbitwallet__pb2.RelayTransferTokenTxnReq.FromString,
                    response_serializer=qbitwallet__pb2.RelayTxnResp.SerializeToString,
            ),
            'RelayTransferTokenTxnBySlave': grpc.unary_unary_rpc_method_handler(
                    servicer.RelayTransferTokenTxnBySlave,
                    request_deserializer=qbitwallet__pb2.RelayTransferTokenTxnBySlaveReq.FromString,
                    response_serializer=qbitwallet__pb2.RelayTxnResp.SerializeToString,
            ),
            'RelaySlaveTxn': grpc.unary_unary_rpc_method_handler(
                    servicer.RelaySlaveTxn,
                    request_deserializer=qbitwallet__pb2.RelaySlaveTxnReq.FromString,
                    response_serializer=qbitwallet__pb2.RelayTxnResp.SerializeToString,
            ),
            'RelaySlaveTxnBySlave': grpc.unary_unary_rpc_method_handler(
                    servicer.RelaySlaveTxnBySlave,
                    request_deserializer=qbitwallet__pb2.RelaySlaveTxnBySlaveReq.FromString,
                    response_serializer=qbitwallet__pb2.RelayTxnResp.SerializeToString,
            ),
            'ChangePassphrase': grpc.unary_unary_rpc_method_handler(
                    servicer.ChangePassphrase,
                    request_deserializer=qbitwallet__pb2.ChangePassphraseReq.FromString,
                    response_serializer=qbitwallet__pb2.ChangePassphraseResp.SerializeToString,
            ),
            'GetTransactionsByAddress': grpc.unary_unary_rpc_method_handler(
                    servicer.GetTransactionsByAddress,
                    request_deserializer=qbitwallet__pb2.TransactionsByAddressReq.FromString,
                    response_serializer=qbitwallet__pb2.TransactionsByAddressResp.SerializeToString,
            ),
            'GetPaginatedTransactionsByAddress': grpc.unary_unary_rpc_method_handler(
                    servicer.GetPaginatedTransactionsByAddress,
                    request_deserializer=qbitwallet__pb2.PaginatedTransactionsByAddressReq.FromString,
                    response_serializer=qbitwallet__pb2.PaginatedTransactionsByAddressResp.SerializeToString,
            ),
            'GetTransaction': grpc.unary_unary_rpc_method_handler(
                    servicer.GetTransaction,
                    request_deserializer=qbitwallet__pb2.TransactionReq.FromString,
                    response_serializer=qbitwallet__pb2.TransactionResp.SerializeToString,
            ),
            'GetBalance': grpc.unary_unary_rpc_method_handler(
                    servicer.GetBalance,
                    request_deserializer=qbitwallet__pb2.BalanceReq.FromString,
                    response_serializer=qbitwallet__pb2.BalanceResp.SerializeToString,
            ),
            'GetTotalBalance': grpc.unary_unary_rpc_method_handler(
                    servicer.GetTotalBalance,
                    request_deserializer=qbitwallet__pb2.TotalBalanceReq.FromString,
                    response_serializer=qbitwallet__pb2.TotalBalanceResp.SerializeToString,
            ),
            'GetOTS': grpc.unary_unary_rpc_method_handler(
                    servicer.GetOTS,
                    request_deserializer=qbitwallet__pb2.OTSReq.FromString,
                    response_serializer=qbitwallet__pb2.OTSResp.SerializeToString,
            ),
            'GetHeight': grpc.unary_unary_rpc_method_handler(
                    servicer.GetHeight,
                    request_deserializer=qbitwallet__pb2.HeightReq.FromString,
                    response_serializer=qbitwallet__pb2.HeightResp.SerializeToString,
            ),
            'GetBlock': grpc.unary_unary_rpc_method_handler(
                    servicer.GetBlock,
                    request_deserializer=qbitwallet__pb2.BlockReq.FromString,
                    response_serializer=qbitwallet__pb2.BlockResp.SerializeToString,
            ),
            'GetBlockByNumber': grpc.unary_unary_rpc_method_handler(
                    servicer.GetBlockByNumber,
                    request_deserializer=qbitwallet__pb2.BlockByNumberReq.FromString,
                    response_serializer=qbitwallet__pb2.BlockResp.SerializeToString,
            ),
            'GetAddressFromPK': grpc.unary_unary_rpc_method_handler(
                    servicer.GetAddressFromPK,
                    request_deserializer=qbitwallet__pb2.AddressFromPKReq.FromString,
                    response_serializer=qbitwallet__pb2.AddressFromPKResp.SerializeToString,
            ),
            'GetNodeInfo': grpc.unary_unary_rpc_method_handler(
                    servicer.GetNodeInfo,
                    request_deserializer=qbitwallet__pb2.NodeInfoReq.FromString,
                    response_serializer=qbitwallet__pb2.NodeInfoResp.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'qrl.WalletAPI', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))
    server.add_registered_method_handlers('qrl.WalletAPI', rpc_method_handlers)


 # This class is part of an EXPERIMENTAL API.
class WalletAPI(object):
    """//////////////////////////
    //////////////////////////
    //////////////////////////
    ////     API       ///////
    //////////////////////////
    //////////////////////////
    //////////////////////////

    This service describes the Wallet API
    """

    @staticmethod
    def AddNewAddress(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/qrl.WalletAPI/AddNewAddress',
            qbitwallet__pb2.AddNewAddressReq.SerializeToString,
            qbitwallet__pb2.AddNewAddressResp.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def AddNewAddressWithSlaves(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/qrl.WalletAPI/AddNewAddressWithSlaves',
            qbitwallet__pb2.AddNewAddressWithSlavesReq.SerializeToString,
            qbitwallet__pb2.AddNewAddressResp.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def ListAddresses(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/qrl.WalletAPI/ListAddresses',
            qbitwallet__pb2.ListAddressesReq.SerializeToString,
            qbitwallet__pb2.ListAddressesResp.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def RemoveAddress(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/qrl.WalletAPI/RemoveAddress',
            qbitwallet__pb2.RemoveAddressReq.SerializeToString,
            qbitwallet__pb2.RemoveAddressResp.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def IsValidAddress(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/qrl.WalletAPI/IsValidAddress',
            qbitwallet__pb2.ValidAddressReq.SerializeToString,
            qbitwallet__pb2.ValidAddressResp.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def EncryptWallet(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/qrl.WalletAPI/EncryptWallet',
            qbitwallet__pb2.EncryptWalletReq.SerializeToString,
            qbitwallet__pb2.EncryptWalletResp.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def LockWallet(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/qrl.WalletAPI/LockWallet',
            qbitwallet__pb2.LockWalletReq.SerializeToString,
            qbitwallet__pb2.LockWalletResp.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def UnlockWallet(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/qrl.WalletAPI/UnlockWallet',
            qbitwallet__pb2.UnlockWalletReq.SerializeToString,
            qbitwallet__pb2.UnlockWalletResp.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetRecoverySeeds(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/qrl.WalletAPI/GetRecoverySeeds',
            qbitwallet__pb2.GetRecoverySeedsReq.SerializeToString,
            qbitwallet__pb2.GetRecoverySeedsResp.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetWalletInfo(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/qrl.WalletAPI/GetWalletInfo',
            qbitwallet__pb2.GetWalletInfoReq.SerializeToString,
            qbitwallet__pb2.GetWalletInfoResp.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def RelayTransferTxn(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/qrl.WalletAPI/RelayTransferTxn',
            qbitwallet__pb2.RelayTransferTxnReq.SerializeToString,
            qbitwallet__pb2.RelayTxnResp.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def RelayTransferTxnBySlave(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/qrl.WalletAPI/RelayTransferTxnBySlave',
            qbitwallet__pb2.RelayTransferTxnBySlaveReq.SerializeToString,
            qbitwallet__pb2.RelayTxnResp.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def RelayMessageTxn(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/qrl.WalletAPI/RelayMessageTxn',
            qbitwallet__pb2.RelayMessageTxnReq.SerializeToString,
            qbitwallet__pb2.RelayTxnResp.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def RelayMessageTxnBySlave(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/qrl.WalletAPI/RelayMessageTxnBySlave',
            qbitwallet__pb2.RelayMessageTxnBySlaveReq.SerializeToString,
            qbitwallet__pb2.RelayTxnResp.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def RelayTokenTxn(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/qrl.WalletAPI/RelayTokenTxn',
            qbitwallet__pb2.RelayTokenTxnReq.SerializeToString,
            qbitwallet__pb2.RelayTxnResp.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def RelayTokenTxnBySlave(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/qrl.WalletAPI/RelayTokenTxnBySlave',
            qbitwallet__pb2.RelayTokenTxnBySlaveReq.SerializeToString,
            qbitwallet__pb2.RelayTxnResp.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def RelayTransferTokenTxn(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/qrl.WalletAPI/RelayTransferTokenTxn',
            qbitwallet__pb2.RelayTransferTokenTxnReq.SerializeToString,
            qbitwallet__pb2.RelayTxnResp.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def RelayTransferTokenTxnBySlave(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/qrl.WalletAPI/RelayTransferTokenTxnBySlave',
            qbitwallet__pb2.RelayTransferTokenTxnBySlaveReq.SerializeToString,
            qbitwallet__pb2.RelayTxnResp.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def RelaySlaveTxn(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/qrl.WalletAPI/RelaySlaveTxn',
            qbitwallet__pb2.RelaySlaveTxnReq.SerializeToString,
            qbitwallet__pb2.RelayTxnResp.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def RelaySlaveTxnBySlave(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/qrl.WalletAPI/RelaySlaveTxnBySlave',
            qbitwallet__pb2.RelaySlaveTxnBySlaveReq.SerializeToString,
            qbitwallet__pb2.RelayTxnResp.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def ChangePassphrase(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/qrl.WalletAPI/ChangePassphrase',
            qbitwallet__pb2.ChangePassphraseReq.SerializeToString,
            qbitwallet__pb2.ChangePassphraseResp.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetTransactionsByAddress(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/qrl.WalletAPI/GetTransactionsByAddress',
            qbitwallet__pb2.TransactionsByAddressReq.SerializeToString,
            qbitwallet__pb2.TransactionsByAddressResp.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetPaginatedTransactionsByAddress(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/qrl.WalletAPI/GetPaginatedTransactionsByAddress',
            qbitwallet__pb2.PaginatedTransactionsByAddressReq.SerializeToString,
            qbitwallet__pb2.PaginatedTransactionsByAddressResp.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetTransaction(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/qrl.WalletAPI/GetTransaction',
            qbitwallet__pb2.TransactionReq.SerializeToString,
            qbitwallet__pb2.TransactionResp.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetBalance(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/qrl.WalletAPI/GetBalance',
            qbitwallet__pb2.BalanceReq.SerializeToString,
            qbitwallet__pb2.BalanceResp.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetTotalBalance(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/qrl.WalletAPI/GetTotalBalance',
            qbitwallet__pb2.TotalBalanceReq.SerializeToString,
            qbitwallet__pb2.TotalBalanceResp.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetOTS(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/qrl.WalletAPI/GetOTS',
            qbitwallet__pb2.OTSReq.SerializeToString,
            qbitwallet__pb2.OTSResp.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetHeight(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/qrl.WalletAPI/GetHeight',
            qbitwallet__pb2.HeightReq.SerializeToString,
            qbitwallet__pb2.HeightResp.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetBlock(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/qrl.WalletAPI/GetBlock',
            qbitwallet__pb2.BlockReq.SerializeToString,
            qbitwallet__pb2.BlockResp.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetBlockByNumber(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/qrl.WalletAPI/GetBlockByNumber',
            qbitwallet__pb2.BlockByNumberReq.SerializeToString,
            qbitwallet__pb2.BlockResp.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetAddressFromPK(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/qrl.WalletAPI/GetAddressFromPK',
            qbitwallet__pb2.AddressFromPKReq.SerializeToString,
            qbitwallet__pb2.AddressFromPKResp.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetNodeInfo(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/qrl.WalletAPI/GetNodeInfo',
            qbitwallet__pb2.NodeInfoReq.SerializeToString,
            qbitwallet__pb2.NodeInfoResp.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)



# coding=utf-8
# Distributed under the MIT software license, see the accompanying
# file LICENSE or http://www.opensource.org/licenses/mit-license.php.
import os
from enum import Enum
from typing import Callable, Set, List
from ipaddress import IPv4Address

import simplejson as json
from pyqryptonight.pyqryptonight import UInt256ToString
from twisted.internet import reactor

from qbitcoin.core import config
from qbitcoin.core.misc import logger, ntp
from qbitcoin.core.misc.expiring_set import ExpiringSet
from qbitcoin.core.notification.Observable import Observable
from qbitcoin.core.notification.ObservableEvent import ObservableEvent
from qbitcoin.core.p2p.IPMetadata import IPMetadata
from qbitcoin.core.p2p.p2pObserver import P2PBaseObserver
from qbitcoin.core.p2p.p2pprotocol import P2PProtocol
from qbitcoin.generated import qbitlegacy_pb2, qbit_pb2


class P2PPeerManager(P2PBaseObserver):
    class EventType(Enum):
        NO_PEERS = 1

    def __init__(self):
        super().__init__()
        self._ping_callLater = None
        self._disconnect_callLater = None
        self._channels = []

        self._peer_node_status = dict()

        self._known_peers = set()
        self.peers_path = os.path.join(config.user.data_dir,
                                       config.dev.peers_filename)

        self.banned_peers_filename = os.path.join(config.user.wallet_dir, config.dev.banned_peers_filename)
        self._banned_peer_ips = ExpiringSet(expiration_time=config.user.ban_minutes * 60,
                                            filename=self.banned_peers_filename)

        self._observable = Observable(self)
        self._p2p_factory = None

    def register(self, message_type: EventType, func: Callable):
        self._observable.register(message_type, func)

    def set_p2p_factory(self, p2p_factory):
        self._p2p_factory = p2p_factory

    @property
    def known_peer_addresses(self):
        return self._known_peers

    def trusted_peer(self, channel: P2PProtocol):
        if self.is_banned(channel.peer):
            return False

        if channel.valid_message_count < config.dev.trust_min_msgcount:
            return False

        if channel.connection_time < config.dev.trust_min_conntime:
            return False

        return True

    @property
    def trusted_addresses(self):
        ip_public_port_set = set()
        for peer in self._p2p_factory.connections:
            if self.trusted_peer(peer) and peer.public_port != 0:
                ip_public_port_set.add(peer.ip_public_port)
        return ip_public_port_set

    @property
    def peer_node_status(self):
        return self._peer_node_status

    def load_known_peers(self) -> List[str]:
        known_peers = []
        try:
            logger.info('Loading known peers')
            with open(self.peers_path, 'r') as infile:
                known_peers = json.load(infile)
        except Exception as e:
            logger.info("Could not open known_peers list")

        return [IPMetadata.canonical_full_address(fa) for fa in known_peers]

    def save_known_peers(self, known_peers: List[str]):
        tmp = list(known_peers)[:3 * config.user.max_peers_limit]
        config.create_path(config.user.data_dir)
        with open(self.peers_path, 'w') as outfile:
            json.dump(tmp, outfile)

    def load_peer_addresses(self) -> None:
        known_peers = self.load_known_peers()
        self._known_peers = self.combine_peer_lists(known_peers, config.user.peer_list, )
        logger.info('Loaded known peers: %s', self._known_peers)
        self.save_known_peers(self._known_peers)

    def extend_known_peers(self, new_peer_addresses: set) -> None:
        new_addresses = set(new_peer_addresses) - self._known_peers

        if self._p2p_factory is not None:
            self._p2p_factory.connect_peer(new_addresses)

        self._known_peers |= set(new_peer_addresses)
        self.save_known_peers(list(self._known_peers))

    @staticmethod
    def combine_peer_lists(peer_ips, sender_full_addresses: List, check_global=False) -> Set[IPMetadata]:
        tmp_list = list(peer_ips)
        tmp_list.extend(sender_full_addresses)

        answer = set()
        for item in tmp_list:
            try:
                answer.add(IPMetadata.canonical_full_address(item, check_global))
            except:  # noqa
                logger.warning("Invalid Peer Address {}".format(item))

        return answer

    def get_better_difficulty(self, current_cumulative_difficulty):
        best_cumulative_difficulty = int(UInt256ToString(current_cumulative_difficulty))
        local_best = best_cumulative_difficulty
        best_channel = None

        for channel in self._peer_node_status:
            node_chain_state = self._peer_node_status[channel]
            node_cumulative_difficulty = int(UInt256ToString(node_chain_state.cumulative_difficulty))
            if node_cumulative_difficulty > best_cumulative_difficulty:
                best_cumulative_difficulty = node_cumulative_difficulty
                best_channel = channel
        logger.debug('Local Best Diff : %s', local_best)
        logger.debug('Remote Best Diff : %s', best_cumulative_difficulty)
        return best_channel

    def insert_to_last_connected_peer(self, ip_public_port, connected_peer=False):
        known_peers = self.load_known_peers()
        connection_set = set()

        if self._p2p_factory is not None:
            # Prepare set of connected peers
            for conn in self._p2p_factory._peer_connections:
                connection_set.add(conn.ip_public_port)

        # Move the current peer to the last position of connected peers
        # or to the start position of disconnected peers
        try:
            index = 0
            if connected_peer:
                if ip_public_port in known_peers:
                    known_peers.remove(ip_public_port)
            else:
                index = known_peers.index(ip_public_port)
                del known_peers[index]

            while index < len(known_peers):
                if known_peers[index] not in connection_set:
                    break
                index += 1
            known_peers.insert(index, ip_public_port)
            self.save_known_peers(known_peers)
        except ValueError:
            pass

    def remove_channel(self, channel):
        self.insert_to_last_connected_peer(channel.ip_public_port)
        if channel in self._channels:
            self._channels.remove(channel)
        if channel in self._peer_node_status:
            del self._peer_node_status[channel]

    def new_channel(self, channel):
        self._channels.append(channel)
        self._peer_node_status[channel] = qbit_pb2.NodeChainState(block_number=0,
                                                                 header_hash=b'',
                                                                 cumulative_difficulty=b'\x00' * 32,
                                                                 timestamp=ntp.getTime())
        channel.register(qbitlegacy_pb2.LegacyMessage.VE, self.handle_version)
        channel.register(qbitlegacy_pb2.LegacyMessage.PL, self.handle_peer_list)
        channel.register(qbitlegacy_pb2.LegacyMessage.CHAINSTATE, self.handle_chain_state)
        channel.register(qbitlegacy_pb2.LegacyMessage.SYNC, self.handle_sync)
        channel.register(qbitlegacy_pb2.LegacyMessage.P2P_ACK, self.handle_p2p_acknowledgement)

    def _get_version_compatibility(self, version) -> bool:
        # Ignore compatibility test on Testnet
        if config.dev.hard_fork_heights == config.dev.testnet_hard_fork_heights:
            return True

        if self._p2p_factory is None:
            return True
        if self._p2p_factory.chain_height >= config.dev.hard_fork_heights[0]:
            try:
                major_version = version.split(".")[0]
                if int(major_version) < 2:
                    return False
            except Exception:
                # Disabled warning as it is not required and could be annoying
                # if a peer with dirty version is trying to connect with the node
                # logger.warning("Exception while checking version for compatibility")
                return True

        if self._p2p_factory.chain_height >= config.dev.hard_fork_heights[1]:
            try:
                major_version = version.split(".")[0]
                if int(major_version) < 3:
                    return False
            except Exception:
                # Disabled warning as it is not required and could be annoying
                # if a peer with dirty version is trying to connect with the node
                # logger.warning("Exception while checking version for compatibility")
                return True

        hard_fork_2 = config.dev.hard_fork_heights[2] + config.dev.hard_fork_node_disconnect_delay[2]
        if self._p2p_factory.chain_height >= hard_fork_2:
            try:
                major_version = version.split(".")[0]
                if int(major_version) < 4:
                    return False
            except Exception:
                # Disabled warning as it is not required and could be annoying
                # if a peer with dirty version is trying to connect with the node
                # logger.warning("Exception while checking version for compatibility")
                return True

        return True

    def handle_version(self, source, message: qbitlegacy_pb2.LegacyMessage):
        """
        Version
        If version is empty, it sends the version & genesis_prev_headerhash.
        Otherwise, processes the content of data.
        In case of mismatches, it disconnects from the peer
        """
        self._validate_message(message, qbitlegacy_pb2.LegacyMessage.VE)

        if not message.veData.version:
            msg = qbitlegacy_pb2.LegacyMessage(
                func_name=qbitlegacy_pb2.LegacyMessage.VE,
                veData=qbitlegacy_pb2.VEData(version=config.dev.version,
                                            genesis_prev_hash=config.user.genesis_prev_headerhash,
                                            rate_limit=config.user.peer_rate_limit))

            source.send(msg)
            return

        logger.info('%s version: %s | genesis prev_headerhash %s',
                    source.peer.ip,
                    message.veData.version,
                    message.veData.genesis_prev_hash)

        if not self._get_version_compatibility(message.veData.version):
            logger.warning("Disconnecting from Peer %s running incompatible node version %s",
                           source.peer.ip,
                           message.veData.version)
            source.loseConnection()
            self.ban_channel(source)
            return

        source.rate_limit = min(config.user.peer_rate_limit, message.veData.rate_limit)

        if message.veData.genesis_prev_hash != config.user.genesis_prev_headerhash:
            logger.warning('%s genesis_prev_headerhash mismatch', source.peer)
            logger.warning('Expected: %s', config.user.genesis_prev_headerhash)
            logger.warning('Found: %s', message.veData.genesis_prev_hash)
            source.loseConnection()
            self.ban_channel(source)

    def handle_peer_list(self, source, message: qbitlegacy_pb2.LegacyMessage):
        P2PBaseObserver._validate_message(message, qbitlegacy_pb2.LegacyMessage.PL)

        if not config.user.enable_peer_discovery:
            return

        if not message.plData.peer_ips:
            return

        # If public port is invalid, ignore rest of the data
        if not (0 < message.plData.public_port < 65536):
            return

        source.set_public_port(message.plData.public_port)

        self.insert_to_last_connected_peer(source.ip_public_port, True)

        sender_peer = IPMetadata(source.peer.ip, message.plData.public_port)

        # Filter out staker messages before processing peer addresses
        actual_peer_ips = []
        staker_messages = []
        
        for peer_ip in message.plData.peer_ips:
            if peer_ip.startswith("STAKER_REQ:") or peer_ip.startswith("STAKER_LIST:"):
                staker_messages.append(peer_ip)
            else:
                actual_peer_ips.append(peer_ip)

        # Check if peer list contains global ip, if it was sent by peer from a global ip address
        new_peers = self.combine_peer_lists(actual_peer_ips,
                                            [sender_peer.full_address],
                                            check_global=IPv4Address(source.peer.ip).is_global)

        logger.info('%s peers data received: %s', source.peer.ip, new_peers)
        if self._p2p_factory is not None:
            self._p2p_factory.add_new_peers_to_peer_q(new_peers)

        # Handle staker messages separately
        self._handle_staker_messages_in_peer_list(staker_messages, source)

    def _handle_staker_messages_in_peer_list(self, staker_messages: list, source):
        """Handle staker-related messages in peer list"""
        try:
            import json
            
            for staker_message in staker_messages:
                if staker_message.startswith("STAKER_REQ:"):
                    # Handle staker request
                    try:
                        staker_json = staker_message[11:]  # Remove "STAKER_REQ:" prefix
                        staker_data = json.loads(staker_json)
                        if self._p2p_factory and self._p2p_factory._qrl_node:
                            peer_host = source.peer.ip
                            logger.info("Received staker list request from peer %s", peer_host)
                            # Send our staker list back - pass the source connection object
                            reactor.callLater(0.5, self._p2p_factory.send_staker_list_to_peer, source)
                    except json.JSONDecodeError as e:
                        logger.warning("Invalid staker request JSON from peer %s: %s", source.peer.ip, str(e))
                        
                elif staker_message.startswith("STAKER_LIST:"):
                    # Handle staker list response  
                    try:
                        staker_json = staker_message[12:]  # Remove "STAKER_LIST:" prefix
                        staker_data = json.loads(staker_json)
                        if self._p2p_factory and self._p2p_factory._qrl_node:
                            peer_host = source.peer.ip
                            stakers = staker_data.get('stakers', [])
                            logger.info("Received staker list from peer %s: %d stakers", peer_host, len(stakers))
                            
                            # Validate and process stakers with duplicate prevention
                            processed_addresses = set()
                            valid_count = 0
                            duplicate_count = 0
                            invalid_count = 0
                            
                            for staker_info in stakers:
                                try:
                                    # Get staker address for duplicate checking
                                    staker_address = staker_info.get('address', '')
                                    
                                    # Handle different address formats
                                    if isinstance(staker_address, tuple):
                                        logger.warning("Address is tuple format from peer %s: %s", peer_host, staker_address)
                                        if len(staker_address) == 1:
                                            staker_address = staker_address[0]
                                        else:
                                            logger.error("Cannot handle multi-element tuple address from peer %s: %s", peer_host, staker_address)
                                            invalid_count += 1
                                            continue
                                    
                                    # Ensure address is string format
                                    if not isinstance(staker_address, str):
                                        logger.error("Address is not string format from peer %s: %s (%s)", peer_host, staker_address, type(staker_address))
                                        invalid_count += 1
                                        continue
                                    
                                    # Check for duplicates
                                    if staker_address in processed_addresses:
                                        duplicate_count += 1
                                        logger.warning("Skipping duplicate staker from peer %s: %s", peer_host, staker_address)
                                        continue
                                        
                                    processed_addresses.add(staker_address)
                                    
                                    # Basic validation
                                    if not staker_address or not staker_info.get('balance', 0):
                                        invalid_count += 1
                                        logger.warning("Invalid staker data from peer %s: missing address or balance", peer_host)
                                        continue
                                    
                                    # Process the staker - format it properly for our handler
                                    formatted_message = {
                                        'type': 'staker_update',
                                        'action': 'add',
                                        'address': staker_address,
                                        'balance': staker_info.get('balance', 0)
                                    }
                                    self._p2p_factory._qrl_node.staking_manager.handle_peer_staker_message(formatted_message, peer_host)
                                    valid_count += 1
                                    
                                except Exception as e:
                                    invalid_count += 1
                                    logger.warning("Error processing staker from peer %s: %s", peer_host, str(e))
                            
                            logger.info("Staker processing complete for peer %s: %d valid, %d duplicates, %d invalid", 
                                       peer_host, valid_count, duplicate_count, invalid_count)
                    except json.JSONDecodeError as e:
                        logger.warning("Invalid staker list JSON from peer %s: %s", source.peer.ip, str(e))
                            
        except Exception as e:
            logger.error("Error handling staker messages from peer %s: %s", source.peer.ip, str(e))

    def handle_sync(self, source, message: qbitlegacy_pb2.LegacyMessage):
        P2PBaseObserver._validate_message(message, qbitlegacy_pb2.LegacyMessage.SYNC)
        if message.syncData.state == '':
            if source.factory.synced:
                source.send_sync(synced=True)

    @staticmethod
    def send_node_chain_state(dest_channel, node_chain_state: qbit_pb2.NodeChainState):
        # FIXME: Not sure this belongs to peer management
        msg = qbitlegacy_pb2.LegacyMessage(func_name=qbitlegacy_pb2.LegacyMessage.CHAINSTATE,
                                          chainStateData=node_chain_state)
        dest_channel.send(msg)

    def monitor_chain_state(self):
        # FIXME: Not sure this belongs to peer management
        current_timestamp = ntp.getTime()
        for channel in self._channels:
            if channel not in self._peer_node_status:
                channel.loseConnection()
                continue
            delta = current_timestamp - self._peer_node_status[channel].timestamp
            if delta > config.user.chain_state_timeout:
                del self._peer_node_status[channel]
                logger.debug('>>>> No State Update [%18s] %2.2f (TIMEOUT)', channel.peer, delta)
                channel.loseConnection()

    def broadcast_chain_state(self, node_chain_state: qbit_pb2.NodeChainState):
        # FIXME: Not sure this belongs to peer management
        # TODO: Verify/Disconnect problematic channels
        # Ping all channels
        for channel in self._channels:
            self.send_node_chain_state(channel, node_chain_state)

        self._observable.notify(ObservableEvent(self.EventType.NO_PEERS))

    def handle_chain_state(self, source, message: qbitlegacy_pb2.LegacyMessage):
        P2PBaseObserver._validate_message(message, qbitlegacy_pb2.LegacyMessage.CHAINSTATE)

        message.chainStateData.timestamp = ntp.getTime()  # Receiving time

        try:
            UInt256ToString(message.chainStateData.cumulative_difficulty)
        except ValueError:
            logger.warning('Invalid Cumulative Difficulty sent by peer')
            source.loseConnection()
            return

        self._peer_node_status[source] = message.chainStateData

        if not self._get_version_compatibility(message.chainStateData.version):
            logger.warning("Disconnecting from Peer %s running incompatible node version %s",
                           source.peer.ip,
                           message.veData.version)
            source.loseConnection()
            return

    def handle_p2p_acknowledgement(self, source, message: qbitlegacy_pb2.LegacyMessage):
        P2PBaseObserver._validate_message(message, qbitlegacy_pb2.LegacyMessage.P2P_ACK)

        source.bytes_sent -= message.p2pAckData.bytes_processed
        if source.bytes_sent < 0:
            logger.warning('Disconnecting Peer %s', source.peer)
            logger.warning('Reason: negative bytes_sent value')
            logger.warning('bytes_sent %s', source.bytes_sent)
            logger.warning('Ack bytes_processed %s', message.p2pAckData.bytes_processed)
            source.loseConnection()

        source.send_next()

    ####################################################
    ####################################################
    ####################################################
    ####################################################
    def is_banned(self, peer: IPMetadata):
        return peer.ip in self._banned_peer_ips

    def ban_channel(self, channel: P2PProtocol):
        self._banned_peer_ips.add(channel.peer.ip)
        logger.warning('Banned %s', channel.peer.ip)
        channel.loseConnection()

    def get_peers_stat(self) -> list:
        peers_stat = []
        # Copying the list of keys, to avoid any change by other thread
        for source in list(self.peer_node_status.keys()):
            try:
                peer_stat = qbit_pb2.PeerStat(peer_ip=source.peer.ip.encode(),
                                             port=source.peer.port,
                                             node_chain_state=self.peer_node_status[source])
                peers_stat.append(peer_stat)
            except KeyError:
                # Ignore in case the key is deleted by other thread causing KeyError
                continue
        return peers_stat



from pyqrllib.pyqrllib import bin2hstr

from qbitcoin.generated import qbit_pb2
from qbitcoin.core.misc import logger
from qbitcoin.core.StateContainer import StateContainer
from qbitcoin.core.PaginatedData import PaginatedData
from qbitcoin.core.txs.multisig.MultiSigVote import MultiSigVote
from qbitcoin.core.State import State
from qbitcoin.core.FalconHelper import falcon_pk_to_address


class VoteStats:
    def __init__(self, protobuf_block=None):
        self._data = protobuf_block
        if protobuf_block is None:
            self._data = qbit_pb2.VoteStats()

    @property
    def pbdata(self):
        return self._data

    def is_active(self, current_block_number) -> bool:
        return not self.executed and current_block_number <= self.expiry_block_number

    @property
    def multi_sig_address(self):
        return self._data.multi_sig_address

    @property
    def expiry_block_number(self):
        return self._data.expiry_block_number

    @property
    def shared_key(self):
        return self._data.shared_key

    @property
    def signatories(self):
        return self._data.signatories

    @property
    def tx_hashes(self):
        return self._data.tx_hashes

    @property
    def unvotes(self):
        return self._data.unvotes

    @property
    def total_weight(self):
        return self._data.total_weight

    @property
    def executed(self):
        return self._data.executed

    def update_total_weight(self, value, subtract):
        if subtract:
            self._data.total_weight -= value
        else:
            self._data.total_weight += value

    def get_address_index(self, address: bytes):
        for i in range(len(self.signatories)):
            if address == self.signatories[i]:
                return i

        return -1

    def get_unvote_by_address(self, address) -> [bool, int]:
        i = self.get_address_index(address)
        if i != -1:
            return self.unvotes[i], i

        return False, -1

    def get_vote_tx_hash_by_signatory_address(self, address):
        i = self.get_address_index(address)
        return self.tx_hashes[i]

    def apply_vote_stats(self,
                         tx: MultiSigVote,
                         weight: int,
                         state_container: StateContainer) -> bool:
        if state_container.block_number > self.expiry_block_number:
            return False

        i = self.get_address_index(tx.addr_from)

        if i == -1:
            return False

        if tx.unvote == self.unvotes[i]:
            return False

        self._data.tx_hashes[i] = tx.txhash

        if tx.unvote:
            self._data.total_weight -= weight
        else:
            self._data.total_weight += weight

        self._data.unvotes[i] = tx.unvote

        multi_sig_spend = state_container.multi_sig_spend_txs[self.shared_key]
        threshold = state_container.addresses_state[self.multi_sig_address].threshold

        # TODO: return bool response of apply function
        self.apply(state_container,
                   multi_sig_spend,
                   state_container.addresses_state,
                   state_container.paginated_tx_hash,
                   state_container.block_number,
                   threshold)

        return True

    def revert_vote_stats(self,
                          tx: MultiSigVote,
                          weight: int,
                          state_container: StateContainer) -> bool:
        if state_container.block_number > self.expiry_block_number:
            return False

        i = self.get_address_index(tx.addr_from)

        if i == -1:
            return False

        if tx.unvote != self.unvotes[i]:
            return False

        if self._data.tx_hashes[i] != tx.txhash:
            return False

        multi_sig_spend = state_container.multi_sig_spend_txs[self.shared_key]
        threshold = state_container.addresses_state[self.multi_sig_address].threshold

        self.revert(state_container,
                    multi_sig_spend,
                    state_container.addresses_state,
                    state_container.paginated_tx_hash,
                    state_container.block_number,
                    threshold)

        self._data.tx_hashes[i] = tx.prev_tx_hash

        if tx.unvote:
            self._data.total_weight += weight
        else:
            self._data.total_weight -= weight

        self._data.unvotes[i] = not tx.unvote

        return True

    @staticmethod
    def create(multi_sig_address: bytes,
               shared_key: bytes,
               signatories: bytes,
               expiry_block_number: int):
        vote_stats = VoteStats()

        vote_stats._data.multi_sig_address = multi_sig_address
        vote_stats._data.shared_key = shared_key
        vote_stats._data.expiry_block_number = expiry_block_number

        for signatory in signatories:
            vote_stats._data.signatories.append(signatory)
            vote_stats._data.tx_hashes.append(b'')
            vote_stats._data.unvotes.append(True)

        return vote_stats

    def apply(self,
              state_container,
              multi_sig_spend,
              addresses_state: dict,
              paginated_tx_hash: PaginatedData,
              current_block_number: int,
              threshold: int) -> bool:
        # TODO: return False if executed
        if self.executed:
            return True

        if self.total_weight < threshold:
            return False

        if current_block_number > self.expiry_block_number:
            return False

        if multi_sig_spend.total_amount > addresses_state[self.multi_sig_address].balance:
            logger.info("[VoteStats] Insufficient funds to execute Multi Sig Spend")
            logger.info("Multi Sig Spend Amount: %s, Funds Available: %s",
                        multi_sig_spend.total_amount,
                        addresses_state[self.multi_sig_address].balance)
            logger.info("Multi Sig Spend txn hash: %s", bin2hstr(multi_sig_spend.txhash))
            logger.info("Multi Sig Address: %s", bin2hstr(multi_sig_spend.multi_sig_address))
            return False

        addresses_state[self.multi_sig_address].update_balance(state_container,
                                                               multi_sig_spend.total_amount,
                                                               subtract=True)

        addr_from_pk = falcon_pk_to_address(multi_sig_spend.PK)
        for index in range(0, len(multi_sig_spend.addrs_to)):
            addr_to = multi_sig_spend.addrs_to[index]
            address_state = addresses_state[addr_to]
            if addr_to not in (multi_sig_spend.addr_from, addr_from_pk):
                paginated_tx_hash.insert(address_state, multi_sig_spend.txhash)
            address_state.update_balance(state_container, multi_sig_spend.amounts[index])

        self._data.executed = True
        return True

    def revert(self,
               state_container,
               multi_sig_spend,
               addresses_state: dict,
               paginated_tx_hash: PaginatedData,
               current_block_number: int,
               threshold: int) -> bool:
        if not self.executed:
            return True

        if self.total_weight < threshold:
            return False

        if current_block_number > self.expiry_block_number:
            return False
        addresses_state[self.multi_sig_address].update_balance(state_container, multi_sig_spend.total_amount)

        addr_from_pk = bytes(falcon_pk_to_address(multi_sig_spend.PK))
        for index in range(0, len(multi_sig_spend.addrs_to)):
            addr_to = multi_sig_spend.addrs_to[index]
            address_state = addresses_state[addr_to]
            if addr_to not in (multi_sig_spend.addr_from, addr_from_pk):
                paginated_tx_hash.remove(address_state, multi_sig_spend.txhash)
            address_state.update_balance(state_container, multi_sig_spend.amounts[index], subtract=True)

        self._data.executed = False
        return True

    def serialize(self):
        return self._data.SerializeToString()

    @staticmethod
    def deserialize(data):
        pbdata = qbit_pb2.VoteStats()
        pbdata.ParseFromString(bytes(data))
        return VoteStats(pbdata)

    def put_state(self, state: State, batch):
        try:
            state._db.put_raw(b'shared_key_' + self.shared_key, self.serialize(), batch)
        except Exception as e:
            raise Exception("[put_state] Exception in VoteStats %s", e)

    @staticmethod
    def delete_state(state: State, shared_key: bytes, batch):
        try:
            state._db.delete(b'shared_key_' + shared_key, batch)
        except Exception as e:
            raise Exception("[delete_state] Exception in VoteStats %s", e)

    @staticmethod
    def get_state(state: State, shared_key):
        try:
            data = state._db.get_raw(b'shared_key_' + shared_key)
            return VoteStats.deserialize(data)
        except KeyError:
            logger.debug('[get_state] VoteStats %s not found', bin2hstr(shared_key).encode())
        except Exception as e:
            logger.error('[get_state] %s', e)

        return None

    # @staticmethod
    # def apply_and_put(state: State,
    #                   state_container: StateContainer):
    #     for key in state_container.votes_stats:
    #         vote_stats = state_container.votes_stats[key]
    #         multi_sig_spend = state_container.multi_sig_spend_txs[vote_stats.shared_key]
    #         threshold = state_container.addresses_state[vote_stats.multi_sig_address].threshold
    #
    #         vote_stats.apply(state_container,
    #                          multi_sig_spend,
    #                          state_container.addresses_state,
    #                          state_container.paginated_tx_hash,
    #                          state_container.block_number,
    #                          threshold)
    #         vote_stats.put_state(state, state_container.batch)
    #
    #     return True
    #
    # @staticmethod
    # def revert_and_put(state: State,
    #                    state_container: StateContainer):
    #     for key in state_container.votes_stats:
    #         vote_stats = state_container.votes_stats[key]
    #         multi_sig_spend = state_container.multi_sig_spend_txs[vote_stats.shared_key]
    #         threshold = state_container.addresses_state[vote_stats.multi_sig_address].threshold
    #
    #         vote_stats.revert(state_container,
    #                           multi_sig_spend,
    #                           state_container.addresses_state,
    #                           state_container.paginated_tx_hash,
    #                           state_container.block_number,
    #                           threshold)
    #         vote_stats.put_state(state, state_container.batch)
    #
    #     return True

    @staticmethod
    def put_all(state: State,
                state_container: StateContainer):
        for key in state_container.votes_stats:
            vote_stats = state_container.votes_stats[key]
            vote_stats.put_state(state, state_container.batch)

        return True

    @staticmethod
    def revert_all(state: State,
                   state_container: StateContainer):
        for key in state_container.votes_stats:
            vote_stats = state_container.votes_stats[key]
            vote_stats.put_state(state, state_container.batch)

        return True



# coding=utf-8
# Distributed under the MIT software license, see the accompanying
# file LICENSE or http://www.opensource.org/licenses/mit-license.php.

from qbitcoin.core import config
from qbitcoin.core.misc import ntp
from qbitcoin.core.txs.Transaction import Transaction


class TransactionInfo:

    def __init__(self, tx: Transaction, block_number: int, timestamp: int=None):
        self._transaction = tx
        self._block_number = block_number
        self._timestamp = timestamp
        if not self._timestamp:
            self._timestamp = ntp.getTime()

    def __lt__(self, tx_info):
        if self.transaction.fee < tx_info.transaction.fee:
            return True

        return False

    @property
    def transaction(self):
        return self._transaction

    @property
    def block_number(self):
        return self._block_number

    @property
    def timestamp(self):
        return self._timestamp

    def is_stale(self, current_block_number: int):
        if current_block_number > self._block_number + config.user.stale_transaction_threshold:
            return True

        # If chain recovered from a fork where chain height is reduced
        # then update block_number of the transactions in pool
        if current_block_number < self._block_number:
            self.update_block_number(current_block_number)

        return False

    def update_block_number(self, current_block_number: int):
        self._block_number = current_block_number

    def validate(self, new_state_container, update_state_container, block_number) -> bool:
        addresses_set = set()
        self.transaction.set_affected_address(addresses_set)
        state_container = new_state_container(addresses_set,
                                              block_number,
                                              False,
                                              None)

        if not update_state_container(self.transaction, state_container):
            return False

        # Nonce should not be checked during transaction validation,
        # as the appropriate nonce can be set by miner before placing
        # the txn into block
        if not self.transaction.validate_all(state_container, False):
            return False
        return True



# coding=utf-8
# Distributed under the MIT software license, see the accompanying
# file LICENSE or http://www.opensource.org/licenses/mit-license.php.

from qbitcoin.core.misc import logger
from qbitcoin.core.Block import Block
from qbitcoin.core.State import State
from qbitcoin.core.txs.Transaction import Transaction
from qbitcoin.core.txs.CoinBase import CoinBase
from qbitcoin.generated import qbitstateinfo_pb2


class LastTransactions:
    def __init__(self, pbdata=None):
        self._data = pbdata
        if not pbdata:
            self._data = qbitstateinfo_pb2.LastTransactions()

    @property
    def tx_metadata(self):
        return self._data.tx_metadata

    def add(self, tx: Transaction, block_number: int, timestamp: int):
        tm = qbitstateinfo_pb2.TransactionMetadata(transaction=tx.pbdata,
                                                  block_number=block_number,
                                                  timestamp=timestamp)
        tmp = self._data.tx_metadata[::-1]
        tmp.append(tm)
        del self._data.tx_metadata[:]
        self._data.tx_metadata.extend(tmp[-20:][::-1])

    def serialize(self) -> str:
        return self._data.SerializeToString()

    @staticmethod
    def deserialize(data):
        pbdata = qbitstateinfo_pb2.LastTransactions()
        pbdata.ParseFromString(bytes(data))
        return LastTransactions(pbdata)

    @staticmethod
    def _remove_last_tx(state: State, block: Block, batch):
        if len(block.transactions) == 0:
            return

        try:
            last_txn = LastTransactions.deserialize(state._db.get_raw(b'last_txn'))
        except KeyError:
            return

        for protobuf_txn in block.transactions:
            txn = Transaction.from_pbdata(protobuf_txn)
            i = 0
            while i < len(last_txn.tx_metadata):
                tx = Transaction.from_pbdata(last_txn.tx_metadata[i].transaction)
                if txn.txhash == tx.txhash:
                    del last_txn.tx_metadata[i]
                    break
                i += 1

        state._db.put_raw(b'last_txn', last_txn.serialize(), batch)

    @staticmethod
    def _update_last_tx(state: State, block: Block, batch):
        if len(block.transactions) == 0:
            return
        last_txn = LastTransactions()

        try:
            last_txn = LastTransactions.deserialize(state._db.get_raw(b'last_txn'))
        except KeyError:
            pass

        for protobuf_txn in block.transactions[-20:]:
            txn = Transaction.from_pbdata(protobuf_txn)
            if isinstance(txn, CoinBase):
                continue
            last_txn.add(txn, block.block_number, block.timestamp)

        state._db.put_raw(b'last_txn', last_txn.serialize(), batch)

    @staticmethod
    def get_last_txs(state: State):
        try:
            last_txn = LastTransactions.deserialize(state._db.get_raw(b'last_txn'))
        except KeyError:
            return []
        except Exception as e:  # noqa
            logger.warning("[get_last_txs] Exception during call %s", e)
            return []

        txs = []
        for tx_metadata in last_txn.tx_metadata:
            data = tx_metadata.transaction
            tx = Transaction.from_pbdata(data)
            txs.append(tx)

        return txs



"""
Falcon post-quantum signature scheme implementation for Qbitcoin.
This module provides secure signing and verification using the Falcon-512 algorithm.
"""

import hashlib
import base64
import logging
from typing import Tuple, Optional, Dict, Any, Union

# Configure logger
logger = logging.getLogger(__name__)

# Import the Falcon 512 module from pqcrypto
from pqcrypto.sign.falcon_512 import (
    generate_keypair,
    sign as pq_sign,
    verify as pq_verify,
    PUBLIC_KEY_SIZE,
    SECRET_KEY_SIZE,
    SIGNATURE_SIZE,
    __ffi as ffi,  # Direct CFFI access
    __lib as lib   # Direct CFFI access
)


class FalconSignature:
    """
    Implementation of the Falcon post-quantum signature scheme.
    Provides methods for key generation, signing messages, and verifying signatures.
    """
    
    ALGORITHM_NAME = "falcon-512"
    
    @classmethod
    def get_algorithm_details(cls) -> Dict[str, int]:
        """
        Get the algorithm details including key and signature sizes
        
        Returns:
            Dict[str, int]: Dictionary with algorithm details
        """
        return {
            "name": cls.ALGORITHM_NAME,
            "public_key_size": PUBLIC_KEY_SIZE,
            "secret_key_size": SECRET_KEY_SIZE,
            "signature_size": SIGNATURE_SIZE
        }
    
    @staticmethod
    def generate_keypair() -> Tuple[bytes, bytes]:
        """
        Generate a new Falcon-512 key pair
        
        Returns:
            Tuple[bytes, bytes]: A tuple containing the private and public keys as bytes
        """
        # Generate the keypair using pqcrypto library
        key1, key2 = generate_keypair()
        
        # Check key sizes to determine which is which
        # In Falcon 512, SECRET_KEY_SIZE is 1281 and PUBLIC_KEY_SIZE is 897
        if len(key1) == SECRET_KEY_SIZE and len(key2) == PUBLIC_KEY_SIZE:
            # Keys are in correct order (secret_key, public_key)
            return key1, key2
        elif len(key1) == PUBLIC_KEY_SIZE and len(key2) == SECRET_KEY_SIZE:
            # Keys are in reverse order (public_key, secret_key)
            return key2, key1
        else:
            # If sizes don't match expected values, raise an error
            raise ValueError(f"Invalid key sizes: {len(key1)}, {len(key2)}. Expected: {SECRET_KEY_SIZE}, {PUBLIC_KEY_SIZE}")
    
    @staticmethod
    def sign_message(message: Union[str, bytes], private_key: bytes) -> bytes:
        """
        Sign a message using the Falcon-512 private key
        
        Args:
            message (Union[str, bytes]): Message to sign, can be string or bytes
            private_key (bytes): Private key for signing
            
        Returns:
            bytes: Signature
        """
        # Convert string message to bytes if necessary
        if isinstance(message, str):
            message = message.encode('utf-8')
        
        # Verify the private key is the correct size
        if len(private_key) != SECRET_KEY_SIZE:
            raise ValueError(f"Private key must be exactly {SECRET_KEY_SIZE} bytes, got {len(private_key)}")
        
        # Use direct CFFI approach for signing
        signature_buf = ffi.new(f"uint8_t [{lib.PQCLEAN_FALCON512_CLEAN_CRYPTO_BYTES}]")
        signature_len_ptr = ffi.new("size_t *", lib.PQCLEAN_FALCON512_CLEAN_CRYPTO_BYTES)
        
        result = lib.PQCLEAN_FALCON512_CLEAN_crypto_sign_signature(
            signature_buf, signature_len_ptr, message, len(message), private_key
        )
        
        if result != 0:
            raise RuntimeError("Signature generation failed")
        
        # Extract signature to bytes
        signature_len = signature_len_ptr[0]
        signature = bytes(ffi.buffer(signature_buf, signature_len))
        
        return signature
    
    @staticmethod
    def sign_message_with_prehash(message: Union[str, bytes], private_key: bytes) -> bytes:
        """
        Sign a message with SHA-256 prehashing for large messages
        
        Args:
            message (Union[str, bytes]): Message to sign, can be string or bytes
            private_key (bytes): Private key for signing
            
        Returns:
            bytes: Signature
        """
        # Convert string message to bytes if necessary
        if isinstance(message, str):
            message = message.encode('utf-8')
        
        # Hash the message first (recommended for large messages)
        message_hash = hashlib.sha256(message).digest()
        
        # Use the regular signing method with the hashed message
        return FalconSignature.sign_message(message_hash, private_key)
    
    @staticmethod
    def verify_signature(message: Union[str, bytes], signature: bytes, public_key: bytes) -> bool:
        """
        Verify a signature using the Falcon-512 public key
        
        Args:
            message (Union[str, bytes]): Original message
            signature (bytes): Signature to verify
            public_key (bytes): Public key for verification
            
        Returns:
            bool: True if signature is valid, False otherwise
        """
        # Convert string message to bytes if necessary
        if isinstance(message, str):
            message = message.encode('utf-8')
        
        # Verify the public key is the correct size
        if len(public_key) != PUBLIC_KEY_SIZE:
            raise ValueError(f"Public key must be exactly {PUBLIC_KEY_SIZE} bytes, got {len(public_key)}")
        
        # Use direct CFFI approach for verification
        try:
            result = lib.PQCLEAN_FALCON512_CLEAN_crypto_sign_verify(
                signature, len(signature), message, len(message), public_key
            )
            return result == 0
        except Exception:
            return False
    
    @staticmethod
    def verify_signature_with_prehash(message: Union[str, bytes], signature: bytes, public_key: bytes) -> bool:
        """
        Verify a signature that was created with prehashing
        
        Args:
            message (Union[str, bytes]): Original message
            signature (bytes): Signature to verify
            public_key (bytes): Public key for verification
            
        Returns:
            bool: True if signature is valid, False otherwise
        """
        # Convert string message to bytes if necessary
        if isinstance(message, str):
            message = message.encode('utf-8')
        
        # Hash the message first (same as in signing)
        message_hash = hashlib.sha256(message).digest()
        
        # Use the regular verification method with the hashed message
        return FalconSignature.verify_signature(message_hash, signature, public_key)
    
    @staticmethod
    def hex_to_bytes(hex_string: str) -> bytes:
        """
        Convert a hex string to bytes
        
        Args:
            hex_string (str): Hex-encoded string
            
        Returns:
            bytes: Decoded bytes
        """
        return bytes.fromhex(hex_string)
    
    @staticmethod
    def bytes_to_hex(byte_data: bytes) -> str:
        """
        Convert bytes to a hex string
        
        Args:
            byte_data (bytes): Bytes to encode
            
        Returns:
            str: Hex-encoded string
        """
        return byte_data.hex()
    
    @classmethod
    def create_signature_object(cls, message: Union[str, bytes], signature: bytes, 
                              public_key: bytes, use_prehash: bool = False) -> Dict[str, Any]:
        """
        Create a complete signature object with all relevant information
        
        Args:
            message (Union[str, bytes]): The message that was signed
            signature (bytes): The signature
            public_key (bytes): The public key used for verification
            use_prehash (bool): Whether prehashing was used
            
        Returns:
            Dict[str, Any]: Complete signature object
        """
        # Convert message to bytes if it's a string
        if isinstance(message, str):
            message_bytes = message.encode('utf-8')
        else:
            message_bytes = message
            
        # Verify public key size
        if len(public_key) != PUBLIC_KEY_SIZE:
            raise ValueError(f"Public key must be exactly {PUBLIC_KEY_SIZE} bytes, got {len(public_key)}")
        
        # Create signature object
        signature_object = {
            "algorithm": cls.ALGORITHM_NAME,
            "message_hash": hashlib.sha256(message_bytes).hexdigest(),
            "signature": cls.bytes_to_hex(signature),
            "public_key": cls.bytes_to_hex(public_key),
            "signature_size": len(signature),
            "use_prehash": use_prehash,
            "is_valid": cls.verify_signature_with_prehash(message_bytes, signature, public_key) 
                        if use_prehash else cls.verify_signature(message_bytes, signature, public_key)
        }
        
        return signature_object

# Top-level wrapper functions for easier use in transaction module

def generate_keys() -> Tuple[str, str]:
    """
    Generate a new keypair for signing
    
    Returns:
        Tuple[str, str]: (private_key, public_key) as base64 strings
    """
    # Generate keys using the FalconSignature class
    private_key_bytes, public_key_bytes = FalconSignature.generate_keypair()
    
    # Convert to base64 for string representation
    private_key = base64.b64encode(private_key_bytes).decode('utf-8')
    public_key = base64.b64encode(public_key_bytes).decode('utf-8')
    
    return private_key, public_key

def create_signature(message: Union[str, bytes], private_key: str) -> str:
    """
    Create a signature for a message
    
    Args:
        message: Message to sign (string or bytes)
        private_key: Private key as base64 string
        
    Returns:
        Signature as base64 string
    """
    # Decode private key from base64
    try:
        # Remove any whitespace and normalize padding
        private_key = private_key.strip()
        # Remove existing padding if any
        private_key = private_key.rstrip('=')
        # Add correct padding
        padding_needed = len(private_key) % 4
        if padding_needed:
            private_key += '=' * (4 - padding_needed)
            
        private_key_bytes = base64.b64decode(private_key)
        
        # Verify the private key is the correct size after decoding
        if len(private_key_bytes) != SECRET_KEY_SIZE:
            raise ValueError(f"Private key must be exactly {SECRET_KEY_SIZE} bytes, got {len(private_key_bytes)}")
        
        # Sign the message
        signature_bytes = FalconSignature.sign_message(message, private_key_bytes)
        
        # Return base64 encoded signature
        return base64.b64encode(signature_bytes).decode('utf-8')
        
    except Exception as e:
        # Log the error but don't expose key details
        logger.error(f"Error in create_signature: {str(e)}")
        raise

def verify_signature(message: Union[str, bytes], signature: str, public_key: str) -> bool:
    """
    Verify a message signature
    
    Args:
        message: Original message (string or bytes)
        signature: Signature as base64 or hex string
        public_key: Public key as base64 or hex string
        
    Returns:
        True if signature is valid, False otherwise
    """
    try:
        # First try to decode public key as hex (as used in wallet files)
        try:
            # Try with hex-encoded public key first (the format used in wallets)
            public_key_bytes = bytes.fromhex(public_key)
        except ValueError:
            # If hex decoding fails, try base64 with proper padding handling
            # Handle public key padding for base64 format
            public_key = public_key.strip()
            public_key = public_key.rstrip('=')
            padding_needed = len(public_key) % 4
            if padding_needed:
                public_key += '=' * (4 - padding_needed)
            public_key_bytes = base64.b64decode(public_key)
        
        # Similarly handle signature - first try hex, then base64
        try:
            # Try with hex-encoded signature
            signature_bytes = bytes.fromhex(signature)
        except ValueError:
            # If hex decoding fails, try base64 with proper padding handling
            signature = signature.strip()
            signature = signature.rstrip('=')
            padding_needed = len(signature) % 4
            if padding_needed:
                signature += '=' * (4 - padding_needed)
            signature_bytes = base64.b64decode(signature)
        
        # Verify the public key size
        if len(public_key_bytes) != PUBLIC_KEY_SIZE:
            raise ValueError(f"Public key must be exactly {PUBLIC_KEY_SIZE} bytes, got {len(public_key_bytes)}")
        
        # Verify the signature
        return FalconSignature.verify_signature(message, signature_bytes, public_key_bytes)
    except Exception as e:
        logger.error(f"Error verifying signature: {e}")
        return False


# coding=utf-8
# Distributed under the MIT software license, see the accompanying
# file LICENSE or http://www.opensource.org/licenses/mit-license.php.
import shutil
import decimal
from os.path import expanduser
from pyqrllib.pyqrllib import hstr2bin
from qbitcoin import __version__ as version
from qbitcoin.generated import qbit_pb2

import os

import yaml
from math import ceil, log


class UserConfig(object):
    __instance = None

    def __init__(self, ignore_check=False):
        # TODO: Move to metaclass in Python 3
        if not ignore_check and UserConfig.__instance is not None:
            raise Exception("UserConfig can only be instantiated once")

        UserConfig.__instance = self

        self.genesis_prev_headerhash = b'The sleeper must awaken'
        self.genesis_timestamp = 1748476800  # Updated to current time (May 29, 2025)
        self.genesis_difficulty = 5000 

        # Default configuration
        self.mining_enabled = False
        self.mining_address = ''
        self.mining_thread_count = 0  # 0 to auto detect thread count based on CPU/GPU number of processors
        self.mining_pause = 0  # this will force a sleep (ms) while mining to reduce cpu usage. Only for mocknet

        # Ephemeral Configuration
        self.accept_ephemeral = True

        # PEER Configuration
        self.max_redundant_connections = 5  # Number of connections allowed from nodes having same IP
        self.enable_peer_discovery = True  # Allows to discover new peers from the connected peers

        self.peer_list = ["134.122.79.166",
                          "220.158.73.254",
                           "98.84.96.236"]

        self.p2p_local_port = 19000  # Locally binded port at which node will listen for connection
        self.p2p_public_port = 19000  # Public port forwarding connections to server

        self.peer_rate_limit = 500  # Max Number of messages per minute per peer
        self.p2p_q_size = 10000
        self.outgoing_message_expiry = 90  # Outgoing message expires after 90 seconds

        self.ntp_servers = ['pool.ntp.org', 'ntp.ubuntu.com']
        self.ntp_refresh = 12 * 60 * 60  # 12 hours
        self.ntp_request_timeout = 10  # 10 seconds ntp timeout
        self.ban_minutes = 20  # Allows to ban a peer's IP who is breaking protocol

        self.monitor_connections_interval = 30  # Monitor connection every 30 seconds
        self.max_peers_limit = 100  # Number of allowed peers
        self.chain_state_timeout = 180
        self.chain_state_broadcast_period = 30
        # must be less than ping_timeout

        self.transaction_minimum_fee = int(0.0001 * (10 ** 9))  # 0.0001 Qbitcoin in quarks
        self.transaction_pool_size = 25000
        self.pending_transaction_pool_size = 75000
        # 1% of the pending_transaction_pool will be reserved for moving stale txn
        self.pending_transaction_pool_reserve = int(self.pending_transaction_pool_size * 0.01)
        self.stale_transaction_threshold = 15  # 15 Blocks

        # Store data in user's application data directory (outside project)
        self._qrl_dir = self._get_default_data_directory()

        # ======================================
        #        ADMIN API CONFIGURATION
        # ======================================
        self.admin_api_enabled = False
        self.admin_api_host = "127.0.0.1"
        self.admin_api_port = 19008
        self.admin_api_threads = 1
        self.admin_api_max_concurrent_rpc = 100

        # ======================================
        #        PUBLIC API CONFIGURATION
        # ======================================
        self.public_api_enabled = True
        self.public_api_host = "0.0.0.0" # only if you want to expose the API to the public 
        self.public_api_port = 19009
        self.public_api_threads = 1
        self.public_api_max_concurrent_rpc = 100

        # ======================================
        #        MINING API CONFIGURATION
        # ======================================
        self.mining_api_enabled = False
        self.mining_api_host = "127.0.0.1"
        self.mining_api_port = 19007
        self.mining_api_threads = 1
        self.mining_api_max_concurrent_rpc = 100

        # ======================================
        #        DEBUG API CONFIGURATION
        # ======================================
        self.debug_api_enabled = False
        self.debug_api_host = "127.0.0.1"
        self.debug_api_port = 52134
        self.debug_api_threads = 1
        self.debug_api_max_concurrent_rpc = 100

        # ======================================
        #        GRPC PROXY CONFIGURATION
        # ======================================
        self.grpc_proxy_host = "127.0.0.1"
        self.grpc_proxy_port = 18090

        # ======================================
        #      WALLET DAEMON CONFIGURATION
        # ======================================
        self.public_api_server = "127.0.0.1:19009"
        self.wallet_daemon_host = "127.0.0.1"
        self.wallet_daemon_port = 18091
        self.number_of_slaves = 3

        # ======================================
        #        WALLET API CONFIGURATION
        # ======================================
        self.wallet_api_host = "0.0.0.0" #same as public api 
        self.wallet_api_port = 19010
        self.wallet_api_threads = 1
        self.wallet_api_max_concurrent_rpc = 100

        # WARNING! loading should be the last line.. any new setting after this will not be updated by the config file
        self.load_yaml(self.config_path)
        # WARNING! loading should be the last line.. any new setting after this will not be updated by the config file

    def _get_default_data_directory(self):
        """Get the default data directory based on the operating system"""
        import platform
        
        system = platform.system()
        home = expanduser("~")
        
        if system == "Windows":
            # Windows: %APPDATA%\Qbitcoin
            app_data = os.environ.get('APPDATA', os.path.join(home, 'AppData', 'Roaming'))
            return os.path.join(app_data, 'Qbitcoin')
        elif system == "Darwin":  # macOS
            # macOS: ~/Library/Application Support/Qbitcoin
            return os.path.join(home, 'Library', 'Application Support', 'Qbitcoin')
        else:  # Linux and other Unix-like systems
            # Linux: ~/.local/share/qbitcoin or ~/.qbitcoin (fallback)
            xdg_data_home = os.environ.get('XDG_DATA_HOME', os.path.join(home, '.local', 'share'))
            return os.path.join(xdg_data_home, 'qbitcoin')

    @property
    def qrl_dir(self):
        return self._qrl_dir

    @qrl_dir.setter
    def qrl_dir(self, new_qrl_dir):
        self._qrl_dir = new_qrl_dir
        self.load_yaml(self.config_path)

    @property
    def wallet_dir(self):
        return expanduser(self.qrl_dir)

    @property
    def data_dir(self):
        return expanduser(os.path.join(self.qrl_dir, "data"))

    @property
    def config_path(self):
        return expanduser(os.path.join(self.qrl_dir, "config.yml"))

    @property
    def log_path(self):
        return expanduser(os.path.join(self.qrl_dir, "qrl.log"))

    @property
    def walletd_log_path(self):
        return expanduser(os.path.join(self.qrl_dir, "walletd.log"))

    @property
    def mining_pool_payment_wallet_path(self):
        return expanduser(os.path.join(self.qrl_dir, 'payment_slaves.json'))

    @staticmethod
    def getInstance():
        if UserConfig.__instance is None:
            return UserConfig()
        return UserConfig.__instance

    def load_yaml(self, file_path):
        """
        Overrides default configuration using a yaml file
        :param file_path: The path to the configuration file
        """
        if os.path.isfile(file_path):
            with open(file_path) as f:
                dataMap = yaml.safe_load(f)
                if dataMap is not None:
                    if 'genesis_prev_headerhash' in dataMap:
                        dataMap['genesis_prev_headerhash'] = dataMap['genesis_prev_headerhash'].encode()
                    self.__dict__.update(**dataMap)


def create_path(path, copy_files=None):
    """Create directory path if it doesn't exist"""
    tmp_path = os.path.join(path)
    if os.path.isdir(tmp_path):
        return
    
    try:
        os.makedirs(tmp_path, exist_ok=True)
        # Import logger here to avoid circular imports
        from qbitcoin.core.misc import logger
        logger.info("Created data directory: %s", tmp_path)
    except OSError as e:
        # Import logger here to avoid circular imports
        from qbitcoin.core.misc import logger
        logger.error("Failed to create data directory %s: %s", tmp_path, str(e))
        raise
    
    if not copy_files:
        return
    for file in copy_files:
        shutil.copy(file, tmp_path)


class DevConfig(object):
    __instance = None

    def __init__(self, pbdata, ignore_check=False, ignore_singleton=False):
        super(DevConfig, self).__init__()
        # TODO: Move to metaclass in Python 3
        if not ignore_check and DevConfig.__instance is not None:
            raise Exception("DevConfig can only be instantiated once")

        if not ignore_singleton:
            DevConfig.__instance = self

        self._data = pbdata

        self.version = version + ' python'

        ################################################################
        # Warning: Don't change following configuration.               #
        #          For Qbitcoin  Developers only                             #
        ################################################################

        self.block_lead_timestamp = 30
        self.block_max_drift = 15
        self.max_future_blocks_length = 256
        self.max_margin_block_number = 32
        self.min_margin_block_number = 7

        self.public_ip = None
        self.cache_frequency = 1000

        self.message_q_size = 300
        self.message_receipt_timeout = 10  # request timeout for full message
        self.message_buffer_size = 64 * 1024 * 1024  # 64 MB

        self.timestamp_error = 5  # Error in second

        self.blocks_per_epoch = 100
         
         

         

        self.default_nonce = 0
        self.default_account_balance = 0 * (10 ** 9)
        self.hash_buffer_size = 4
        self.minimum_minting_delay = 45  # Minimum delay in second before a block is being created

        self.tx_extra_overhead = 15  # 15 bytes
        self.coinbase_address = b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00' \
                                b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'

        # Directories and files
        self.db_name = 'state'
        self.peers_filename = 'known_peers.json'
        self.chain_file_directory = 'data'
        self.wallet_dat_filename = 'wallet.wallet'
        self.slave_dat_filename = 'slave.qrl'
        self.banned_peers_filename = 'banned_peers.qrl'

        self.trust_min_msgcount = 10
        self.trust_min_conntime = 10

        self.supplied_coins = 20000000 * (10 ** 9)  # 20 million QBCfor genesis

        # ======================================
        #      STATE VERSION
        # ======================================
        # Max number of data to be stored per key
        self._state_version = 3

        # ======================================
        #      STATE PAGINATION CONTROLLER
        # ======================================
        # Max number of data to be stored per key
        self._data_per_page = 10000

        # ======================================
        #            P2P SETTINGS
        # ======================================
        self.max_receivable_bytes = 10 * 1024 * 1024  # 10 MB [Temporary Restriction]
        self.reserved_quota = 1024  # 1 KB
        self.max_bytes_out = self.max_receivable_bytes - self.reserved_quota
        self.sync_delay_mining = 60  # Delay mining by 60 seconds while syncing blocks to mainchain

        # ======================================
        #            API SETTINGS
        # ======================================
        self.block_timeseries_size = 1440

        # ======================================
        # SHOR PER QUANTA / MAX ALLOWED DECIMALS
        # ======================================
        self.quark_per_qbitcoin = decimal.Decimal(10 ** 9)

        # ======================================
        # # FOUNDATION MULTI-SIG ADDRESSES
        # # ======================================
        # self.percentage = 100  # multiplier for foundation_address_threshold_percentage to keep it an integer
        # self.foundation_multi_sig_addresses = [b'\x11\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
        #                                        b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
        #                                        b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00',
        #                                        ]

        # ======================================
        # HARD FORK HEIGHTS LIST
        # ======================================
        
        self.hard_fork_heights = [63000, 0, 0]
        self.hard_fork_node_disconnect_delay = [0, 0, 0]
        self.testnet_hard_fork_heights = [63000, 0, 0]
        self.testnet_hard_fork_node_disconnect_delay = [0, 0, 0]

        # ======================================
        # PROPOSAL CONFIG
        # ======================================
        self.proposal_unit_percentage = 100
        self.banned_address = [bytes(hstr2bin('0140ce725b1be80b7314e95fa72be3f0c89af9c2d8a2371cd9'))]

        # ======================================
        # STAKING CONFIG
        # ======================================
        self.staking_enabled = True
        self.staking_reward_interval = 60 # blocks
        self.staking_reward_amount = 180000000000  # 180 QBC in Quarks (9 decimals)
        self.staking_min_balance = 3500000000000  # 3500 QBC minimum to stake (9 decimals)
        self.staking_online_check_interval = 10  # seconds
        self.staking_max_offline_time = 300  # 5 minutes max offline time
        self.staking_max_outputs_per_tx = 100  # Maximum outputs per reward transaction

    @property
    def pbdata(self):
        return self._data

    @property
    def state_version(self):
        return self._state_version

    @property
    def prev_state_key(self):
        return self._data.prev_state_key

    @property
    def current_state_key(self):
        return self._data.current_state_key

    @property
    def activation_header_hash(self):
        return self._data.activation_header_hash

    @property
    def activation_block_number(self):
        return self._data.activation_block_number

    @property
    def data_per_page(self):
        return self._data_per_page

    #@property
   # def ots_tracking_per_page(self):
       # return self._ots_tracking_per_page

    #@property
    #def ots_bitfield_size(self):
        #return self._ots_bitfield_size

    @property
    def reorg_limit(self):
        return self.pbdata.chain.reorg_limit

    @property
    def max_coin_supply(self):
        return decimal.Decimal(self.pbdata.chain.max_coin_supply)

    @property
    def complete_emission_time_span_in_years(self):
        return self.pbdata.chain.complete_emission_time_span_in_years

    @property
    def coin_remaining_at_genesis(self):
        # For 30M max supply with 20M allocated at genesis, 10M remains for mining rewards
        return decimal.Decimal(self.pbdata.chain.max_coin_supply - 20000000)

    @property
    def mining_nonce_offset(self):
        return self.pbdata.block.mining_nonce_offset

    @property
    def extra_nonce_offset(self):
        return self.pbdata.block.extra_nonce_offset

    @property
    def mining_blob_size_in_bytes(self):
        return self.pbdata.block.mining_blob_size_in_bytes

    @property
    def block_timing_in_seconds(self):
        return self.pbdata.block.block_timing_in_seconds

    @property
    def number_of_blocks_to_analyze(self):
        return self.pbdata.block.block_size_controller.number_of_blocks_analyze

    @property
    def size_multiplier(self):
        return self.pbdata.block.block_size_controller.size_multiplier / 100

    @property
    def block_min_size_limit_in_bytes(self):
        return self.pbdata.block.block_size_controller.block_min_size_limit_in_bytes

    @property
    def transaction_multi_output_limit(self):
        return self.pbdata.transaction.multi_output_limit

    # @property
    # def foundation_address_threshold_percentage(self):
    #     return self.pbdata.transaction.foundation_multi_sig.threshold_percentage

    @property
    def message_max_length(self):
        return self.pbdata.transaction.message.max_length

    @property
    def slave_pk_max_length(self):
        return self.pbdata.transaction.slave.slave_pk_max_length

    @property
    def max_token_symbol_length(self):
        return self.pbdata.transaction.token.symbol_max_length

    @property
    def max_token_name_length(self):
        return self.pbdata.transaction.token.name_max_length

    @property
    def lattice_pk1_max_length(self):
        return self.pbdata.transaction.lattice.pk1_max_length

    @property
    def lattice_pk2_max_length(self):
        return self.pbdata.transaction.lattice.pk2_max_length

    @property
    def lattice_pk3_max_length(self):
        return self.pbdata.transaction.lattice.pk3_max_length

    @property
    def proposal_threshold_per(self):
        return self.pbdata.transaction.proposal.threshold_per

    @property
    def default_proposal_options(self):
        return self.pbdata.transaction.proposal.default_options

    @property
    def description_max_length(self):
        return self.pbdata.transaction.proposal.description_max_length

    @property
    def options_max_number(self):
        return self.pbdata.transaction.proposal.options_max_number

    @property
    def option_max_text_length(self):
        return self.pbdata.transaction.proposal.option_max_text_length

    @property
    def proposal_config_activation_delay(self):
        return self.pbdata.transaction.proposal.proposal_config_activation_delay

    @property
    def N_measurement(self):
        return self.pbdata.pow.N_measurement

    @property
    def kp(self):
        return self.pbdata.pow.kp

    @staticmethod
    def getInstance(prev_state_key,
                    current_state_key: bytes,
                    current_block_header_hash: bytes,
                    current_block_number: int):
        if DevConfig.__instance is None:
            return DevConfig.create(prev_state_key,
                                    current_state_key,
                                    current_block_header_hash,
                                    current_block_number)
        return DevConfig.__instance

    @staticmethod
    def get_state_key(headerhash: bytes):
        return b'dev_config_' + headerhash

    @staticmethod
    def create(prev_state_key,
               current_state_key: bytes,
               current_block_header_hash: bytes,
               current_block_number: int,
               ignore_check=False,
               ignore_singleton=False):
        chain = qbit_pb2.DevConfig.Chain(reorg_limit=300,
                                        max_coin_supply=30000000,
                                        complete_emission_time_span_in_years=4)  # Changed to 4 years for halving
        block_size_controller = qbit_pb2.DevConfig.Block.BlockSizeController(number_of_blocks_analyze=10,
                                                                            size_multiplier=110,  # 1.1
                                                                            block_min_size_limit_in_bytes=1024 * 1024)
        block = qbit_pb2.DevConfig.Block(mining_nonce_offset=39,
                                        extra_nonce_offset=43,
                                        mining_blob_size_in_bytes=76,
                                        block_timing_in_seconds=60,  # 1 minute blocks
                                        block_size_controller=block_size_controller)

        message = qbit_pb2.DevConfig.Transaction.Message(max_length=80)

        slave = qbit_pb2.DevConfig.Transaction.Slave(slave_pk_max_length=67)

        token = qbit_pb2.DevConfig.Transaction.Token(symbol_max_length=10,
                                                    name_max_length=30)

        lattice = qbit_pb2.DevConfig.Transaction.Lattice(pk1_max_length=1088,
                                                        pk2_max_length=1472,
                                                        pk3_max_length=65)

        foundation_multi_sig = qbit_pb2.DevConfig.Transaction.FoundationMultiSig(threshold_percentage=10)

        # proposal_config_activation_delay should not be less than reorg_limit
        proposal = qbit_pb2.DevConfig.Transaction.Proposal(threshold_per=51,
                                                          default_options=["YES", "NO", "ABSTAIN"],
                                                          description_max_length=400,
                                                          options_max_number=100,
                                                          option_max_text_length=30,
                                                          proposal_config_activation_delay=10)

        transaction = qbit_pb2.DevConfig.Transaction(multi_output_limit=100,
                                                    message=message,
                                                    slave=slave,
                                                    token=token,
                                                    lattice=lattice,
                                                    foundation_multi_sig=foundation_multi_sig,
                                                    proposal=proposal)

        dev_config = qbit_pb2.DevConfig(chain=chain,
                                       block=block,
                                       transaction=transaction,
                                       pow=qbit_pb2.DevConfig.POW(N_measurement=30, kp=5))

        if prev_state_key is not None:
            dev_config.prev_state_key = prev_state_key
        dev_config.current_state_key = current_state_key

        dev_config.activation_header_hash = current_block_header_hash
        dev_config.activation_block_number = current_block_number

        return DevConfig(dev_config,
                         ignore_check,
                         ignore_singleton)

    def serialize(self) -> str:
        return self._data.SerializeToString()

    @staticmethod
    def deserialize(data):
        pbdata = qbit_pb2.DevConfig()
        pbdata.ParseFromString(bytes(data))
        dev_config = DevConfig(pbdata)
        return dev_config

    def update_from_pbdata(self, pbdata):
        self._data = pbdata


# Hard coded Genesis Header Hash, must be updated if any change is made to genesis block
# Updated Genesis Hash: b2a7aa80ddfa84ac117c5822bd4046160585238f850ffe8265f3ba1d26616264 (July 5, 2025)
genesis_header_hash = bytes(hstr2bin('b2a7aa80ddfa84ac117c5822bd4046160585238f850ffe8265f3ba1d26616264'))
dev = DevConfig.getInstance(None,
                            DevConfig.get_state_key(genesis_header_hash),
                            genesis_header_hash,
                            0)
user = UserConfig.getInstance()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from setuptools import setup, find_packages
from setuptools.command.install import install
import os
import sys


def read_requirements():
    """Read requirements from requirements.txt"""
    requirements_path = os.path.join(os.path.dirname(__file__), 'requirements.txt')
    with open(requirements_path, 'r') as f:
        requirements = []
        for line in f:
            line = line.strip()
            if line and not line.startswith('#') and not line.startswith('pyqrllib') and not line.startswith('pyqryptonight') and not line.startswith('pyqrandomx'):
                requirements.append(line)
        return requirements


def get_version():
    """Get version from the centralized version file"""
    version_file = os.path.join(os.path.dirname(__file__), 'qbitcoin', 'version.py')
    with open(version_file, 'r') as f:
        for line in f:
            if line.startswith('__version__'):
                return line.split('=')[1].strip().strip('"\'')
    return "1.1.3"  # fallback version


class PostInstallCommand(install):
    """Custom post-installation command"""
    def run(self):
        install.run(self)
        
        # Run smart installer for quantum libraries after basic installation
        try:
            print("🔧 Running Qbitcoin smart installer for quantum libraries...")
            print("📋 This will install: pyqrllib, pyqryptonight, pyqrandomx")
            print("⏳ This may take a few minutes for compilation...")
            
            # Import and run smart installer
            from qbitcoin.smart_installer import SmartInstaller
            installer = SmartInstaller()
            
            print("🧬 Installing quantum-resistant libraries with mining support...")
            success = installer.install_all_quantum_libraries()
            
            if success:
                print("✅ Smart installation completed successfully!")
            else:
                print("⚠️  Some quantum libraries may have failed - basic functionality available")
                print("💡 You can manually run: python -m qbitcoin.smart_installer")
            
        except Exception as e:
            print(f"⚠️  Smart installer encountered issues: {e}")
            print("💡 You can manually install quantum libraries by running:")
            print("   python -m qbitcoin.smart_installer")
            # Don't fail the entire installation for quantum library issues


# Read the contents of README file
this_directory = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

# Get requirements from requirements.txt
install_requirements = read_requirements()
quantum_requirements = [
    'pqcrypto>=0.3.0',
]

# Note: Advanced quantum libraries (pyqrllib, pyqryptonight, pyqrandomx) 
# are installed separately by the smart installer to handle compilation issues

setup(
    name='qbitcoin',
    version=get_version(),
    author='Hamza',
    author_email='qbitcoin@example.com',
    description='A Python-based cryptocurrency implementation with quantum-resistant features',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/Hamza1s34/Qbitcoin',
    project_urls={
        'Bug Reports': 'https://github.com/Hamza1s34/Qbitcoin/issues',
        'Source': 'https://github.com/Hamza1s34/Qbitcoin',
        'Documentation': 'https://github.com/Hamza1s34/Qbitcoin/tree/main/docs',
    },
    packages=find_packages(),
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Security :: Cryptography',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
    ],
    python_requires='>=3.8',
    install_requires=install_requirements,
    extras_require={
        'quantum-full': [
            # Note: These require compilation and are better installed via smart installer
            # 'pyqrllib>=1.2.3',
            # 'pyqryptonight>=0.99.0', 
            # 'pyqrandomx>=0.3.0',
        ],
        'dev': [
            'pytest>=7.0.0',
            'pytest-cov>=4.0.0',
            'black>=22.0.0',
            'flake8>=5.0.0',
            'build>=0.8.0',
            'twine>=4.0.0',
        ]
    },
    entry_points={
        'console_scripts': [
            'qbitcoin=qbitcoin.main:main',
            'qbitcoin-node=start_qbitcoin:main_entry',
            'qbitcoin-installer=qbitcoin.smart_installer:main',
        ],
    },
    cmdclass={
        'install': PostInstallCommand,
    },
    include_package_data=True,
    package_data={
        'qbitcoin': ['core/*.yml', '**/*.proto'],
    },
    zip_safe=False,
)



#!/usr/bin/env python3
# coding=utf-8
# Script to create and send a transaction using genesis keys in QRL/Qbitcoin

import os
import sys
import json
import grpc
import time

from pyqrllib.pyqrllib import hstr2bin, bin2hstr

# Add QRL modules to path
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '.')))

from qbitcoin.generated import qbit_pb2, qbit_pb2_grpc
from qbitcoin.core.txs.TransferTransaction import TransferTransaction
from qbitcoin.core.txs.Transaction import Transaction
from qbitcoin.crypto.falcon import FalconSignature
from qbitcoin.tools.wallet_creator import WalletCreator
from qbitcoin.core.AddressState import AddressState
from qbitcoin.core import config

# Constants
NODE_GRPC_ENDPOINT = "localhost:19009"  # Default QRL gRPC endpoint
QUARK_PER_QBITCOIN = 10**9  # 1 Qbitcoin = 10^9 quark (smallest unit)
CONNECTION_TIMEOUT = 5  # seconds

def load_genesis_keys(file_path):
    """Load the genesis keys from the provided JSON file"""
    with open(file_path, 'r') as f:
        genesis_data = json.load(f)
    
    print(f"Loaded genesis address: {genesis_data['address']}")
    
    # Convert hex to bytes
    public_key = bytes.fromhex(genesis_data['public_key_hex'])
    private_key = bytes.fromhex(genesis_data['private_key_hex'])
    
    # Verify the algorithm
    if genesis_data['algorithm'] != "falcon-512":
        raise ValueError(f"Expected falcon-512 algorithm, got {genesis_data['algorithm']}")
    
    return {
        'address': genesis_data['address'],
        'address_bytes': bytes(hstr2bin(genesis_data['address'][1:])),  # Remove the 'Q' prefix
        'public_key': public_key,
        'private_key': private_key
    }

def create_new_wallet():
    """Create a new wallet using Falcon-512"""
    # Generate a new key pair
    private_key, public_key = WalletCreator.create_keypair()
    
    # Use WalletCreator to generate address directly from public key
    address = WalletCreator.generate_address(public_key)
    address_bytes = bytes(hstr2bin(address[1:]))  # Remove 'Q' prefix and convert to bytes
    
    print(f"Created new wallet address: {address}")
    
    return {
        'address': address,
        'address_bytes': address_bytes,
        'public_key': public_key,
        'private_key': private_key
    }

def create_and_sign_transaction(sender, receiver_address, amount_quark):
    """Create and sign a transfer transaction"""
    print(f"Creating transaction: {sender['address']} -> {receiver_address} ({amount_quark} quark)")
    
    # We need to manually handle the transaction since the addr_from is derived incorrectly
    tx = TransferTransaction()
    tx._data.public_key = sender['public_key']
    
    # Add receiver address (remove 'Q' prefix and convert to bytes)
    tx._data.transfer.addrs_to.append(bytes(hstr2bin(receiver_address[1:])))
    
    # Add amount
    tx._data.transfer.amounts.append(amount_quark)
    
    # Set fee
    tx._data.fee = 1000000  # 1 million quark fee
    
    # Important: Override the default Transaction behavior by directly setting addr_from
    # This bypasses the QRLHelper.getAddress() call that fails with Falcon keys
    addr_bytes = sender['address_bytes']
    tx._data.master_addr = addr_bytes
    
    # Get transaction data to sign
    tx_data = tx.get_data_hash()
    
    # Sign with Falcon-512
    print("Signing transaction with genesis key...")
    signature = FalconSignature.sign_message(tx_data, sender['private_key'])
    print(f"DEBUG: Generated signature length: {len(signature)} bytes")
    print(f"DEBUG: Expected max signature size: {FalconSignature.get_algorithm_details()['signature_size']} bytes")
    print(f"DEBUG: Signature hash: {signature[:20].hex()}...")
    tx._data.signature = signature
    
    # Update transaction hash after signing
    tx.update_txhash()
    
    return tx

def validate_transaction(tx):
    """
    Custom validation for transaction before sending to the node
    This bypasses the default validation that uses QRLHelper
    """
    if not tx._data.transfer.addrs_to:
        print("Transaction has no recipient addresses")
        return False
    
    if not tx._data.transfer.amounts:
        print("Transaction has no amounts")
        return False
    
    for amount in tx._data.transfer.amounts:
        if amount <= 0:
            print(f"Invalid amount: {amount}")
            return False
    
    if tx._data.fee < 0:
        print(f"Invalid fee: {tx._data.fee}")
        return False
    
    if not tx._data.signature:
        print("Transaction is not signed")
        return False
    
    return True

def send_transaction(tx):
    """Send a transaction to the QRL node"""
    try:
        # First validate the transaction locally
        if not validate_transaction(tx):
            print("Invalid transaction, cannot submit to node")
            return False
        
        # Set up the gRPC connection
        channel = grpc.insecure_channel(NODE_GRPC_ENDPOINT)
        stub = qbit_pb2_grpc.PublicAPIStub(channel)
        
        # Create the push transaction request
        push_transaction_req = qbit_pb2.PushTransactionReq(transaction_signed=tx.pbdata)
        
        print("Sending transaction to node...")
        push_transaction_resp = stub.PushTransaction(push_transaction_req, timeout=CONNECTION_TIMEOUT)
        
        if push_transaction_resp.error_code == qbit_pb2.PushTransactionResp.SUBMITTED:
            print("Transaction successfully submitted!")
            print(f"Transaction hash: {bin2hstr(tx.txhash)}")
            return True
        else:
            print(f"Transaction submission failed with error: {push_transaction_resp.error_description}")
            return False
            
    except Exception as e:
        print(f"Error sending transaction: {str(e)}")
        return False

def get_user_input():
    """Get user input for recipient address and amount"""
    print("=== Qbitcoin Transaction Creator ===")
    print()
    
    # Get recipient address
    print("Enter recipient address (or press Enter to generate a new wallet):")
    recipient_address = input("Recipient address: ").strip()
    
    # If no address provided, create a new wallet
    if not recipient_address:
        print("\nNo address provided. Generating a new wallet...")
        new_wallet = create_new_wallet()
        recipient_address = new_wallet['address']
        
        # Save the new wallet
        wallet_data = {
            "address": new_wallet['address'],
            "public_key": new_wallet['public_key'].hex(),
            "private_key": new_wallet['private_key'].hex(),
            "algorithm": "falcon-512"
        }
        
        wallet_filename = f"generated_wallet_{int(time.time())}.json"
        with open(wallet_filename, 'w') as f:
            json.dump(wallet_data, f, indent=4)
        print(f"New wallet saved to: {wallet_filename}")
        print(f"Generated recipient address: {recipient_address}")
    else:
        # Validate the provided address
        if not recipient_address.startswith('Q'):
            print("Error: Qbitcoin addresses must start with 'Q'")
            return None, None
        if len(recipient_address) != 51:  # Q + 50 hex chars
            print("Error: Invalid address length. Qbitcoin addresses should be 51 characters long")
            return None, None
        
        # Try to validate the hex format
        try:
            address_bytes = bytes(hstr2bin(recipient_address[1:]))
        except Exception as e:
            print(f"Error: Invalid address format: {e}")
            return None, None
    
    print()
    
    # Get amount to send
    print("Enter amount to send in QBC (or press Enter for default 1000 QBC):")
    amount_input = input("Amount (QBC): ").strip()
    
    if not amount_input:
        amount_qbc = 1000  # Default amount
        print(f"Using default amount: {amount_qbc} QBC")
    else:
        try:
            amount_qbc = float(amount_input)
            if amount_qbc <= 0:
                print("Error: Amount must be positive")
                return None, None
        except ValueError:
            print("Error: Invalid amount format")
            return None, None
    
    # Convert to quark (smallest unit)
    amount_quark = int(amount_qbc * QUARK_PER_QBITCOIN)
    
    print(f"\nTransaction details:")
    print(f"Recipient: {recipient_address}")
    print(f"Amount: {amount_qbc} QBC ({amount_quark} quark)")
    print()
    
    # Confirm transaction
    confirm = input("Proceed with transaction? (y/N): ").strip().lower()
    if confirm not in ['y', 'yes']:
        print("Transaction cancelled.")
        return None, None
    
    return recipient_address, amount_quark

def main():
    try:
        # Get user input
        recipient_address, amount_quark = get_user_input()
        if recipient_address is None or amount_quark is None:
            return
        
        # Path to genesis keys JSON file
        genesis_keys_path = os.path.join(os.path.dirname(__file__), '..', 'genesis_keys.json')
        
        # Check if genesis keys exist
        if not os.path.exists(genesis_keys_path):
            print(f"Error: Genesis keys file not found at {genesis_keys_path}")
            print("Please ensure genesis_keys.json exists in the project root directory.")
            return
        
        # Load genesis keys (sender)
        genesis_keys = load_genesis_keys(genesis_keys_path)
        print(f"Loaded sender address: {genesis_keys['address']}")
        print(f"Genesis public key length: {len(genesis_keys['public_key'])} bytes")
        
        # Create and sign a transaction
        print(f"\nCreating transaction...")
        tx = create_and_sign_transaction(genesis_keys, recipient_address, amount_quark)
        print(f"Transaction created with txhash: {bin2hstr(tx.txhash)}")
        
        # Send the transaction to the node
        print(f"Sending transaction to node...")
        success = send_transaction(tx)
        
        if success:
            print(f"\n✅ Transaction successful!")
            print(f"Amount sent: {amount_quark/QUARK_PER_QBITCOIN} QBC")
            print(f"From: {genesis_keys['address']}")
            print(f"To: {recipient_address}")
            print(f"Transaction hash: {bin2hstr(tx.txhash)}")
        else:
            print("\n❌ Transaction failed.")
            
    except Exception as e:
        print(f"Error in main function: {str(e)}")
        import traceback
        traceback.print_exc()

if __name__ == '__main__':
    main()
